﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: AbstractedLinesFilter.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.OpenGL;
using VideoLib;
using System.ComponentModel;
using OpenTK;

namespace RGBZcamera.Library.AbstractedLines
{
	public class AbstractedLinesFilter : Node, IOpenGLNode
	{
		#region Proxy properties

		[Category("1. Tone renderer"), DisplayName("Light direction"), Description("Position or direction of the single point light source.")]
		public Vector3 LightDirection { get; set; }

		[Category("1. Tone renderer"), DefaultValue(false), DisplayName("Is position?"), Description("If set to true, will treat light position as a direction instead.")]
		public bool IsPosition { get; set; }

		[Category("1. Tone renderer"), DefaultValue(0.0f), Description("The amount of ambient illumination to use in the diffuse and toon shaded images.")]
		public float Ambient { get; set; }

		[Category("1. Tone renderer"), DefaultValue(1.0f), Description("The amount of diffuse illumination to use in the diffuse and toon shaded images.")]
		public float Diffuse { get; set; }

		[Category("1. Tone renderer"), DefaultValue(1.0f), Description("The amount of specular illumination to use in the specular and toon shaded images.")]
		public float Specular { get; set; }

		[Category("1. Tone renderer"), DefaultValue(2.0f), Description("The shininess of the material, the larger the more mirror-like.")]
		public float Shininess { get; set; }

		[Category("1. Tone renderer"), Description("The parameters of the toon step function: x + y * smoothstep(z, w, phong).")]
		public Vector4 ToonStep { get; set; }

		[Category("2. Tone blur"), DefaultValue(8.0f), Description("Standard deviation of the Gaussian blur.")]
		public float Sigma { get; set; }

		[Category("3. Ridge detection"), DefaultValue(8.0f), DisplayName("Line width"), Description("Width of lines to draw, in pixels.")]
		public float LineWidth { get; set; }

		[Category("3. Ridge detection"), DefaultValue(5), DisplayName("Steps"), Description("The maximum number of steps towards the ridge or valley.")]
		public int NumSteps { get; set; }

		[Category("3. Ridge detection"), DefaultValue(0.3f), Description("Fraction of the full distance to move towards line.")]
		public float Beta { get; set; }

		[Category("3. Ridge detection"), DisplayName("Valleys"), Description("Lower and upper curvature threshold, as well as alpha for dark valleys.")]
		public Vector3 Valleys { get; set; }

		[Category("3. Ridge detection"), DisplayName("Ridges (diffuse)"), Description("Lower and upper curvature threshold, as well as alpha for light ridges in the diffuse tone image.")]
		public Vector3 RidgesD { get; set; }

		[Category("3. Ridge detection"), DisplayName("Ridges (specular)"), Description("Lower and upper curvature threshold, as well as alpha for light ridges in the specular tone image.")]
		public Vector3 RidgesS { get; set; }

		[Category("4. Output"), DefaultValue(BlendMode.Multicative), DisplayName("Blend mode"), Description("How to blend toon and video colours.")]
		public BlendMode BlendMode { get; set; }

		[Category("4. Output"), DefaultValue(0.6f), DisplayName("Blend factor"), Description("Factor for blending between input video (0) and toon look (1).")]
		public float BlendFactor { get; set; }

		[Category("4. Output"), DefaultValue(1.4f), Description("Saturation of the video colour.")]
		public float Saturation { get; set; }

		public bool ShouldSerializeLightDirection() { return !LightDirection.Equals(new Vector3(0.2f, 0.2f, -1.0f)); }
		public void ResetLightDirection() { LightDirection = new Vector3(0.2f, 0.2f, -1.0f); }
		public bool ShouldSerializeToonStep() { return !ToonStep.Equals(new Vector4(0.7f, 0.3f, 0.4f, 0.6f)); }
		public void ResetToonStep() { ToonStep = new Vector4(0.7f, 0.3f, 0.4f, 0.6f); }
		public bool ShouldSerializeValleys() { return !Valleys.Equals(new Vector3(0.01f, 0.05f, 1.0f)); }
		public void ResetValleys() { Valleys = new Vector3(0.01f, 0.05f, 1.0f); }
		public bool ShouldSerializeRidgesD() { return !RidgesD.Equals(new Vector3(0.01f, 0.05f, 1.0f)); }
		public void ResetRidgesD() { RidgesD = new Vector3(0.01f, 0.05f, 1.0f); }
		public bool ShouldSerializeRidgesS() { return !RidgesS.Equals(new Vector3(0.01f, 0.05f, 1.0f)); }
		public void ResetRidgesS() { RidgesS = new Vector3(0.01f, 0.05f, 1.0f); }

		#endregion Proxy properties

		private ToneRendererNode toneRenderer;
		private ToneBlurNode toneBlur;
		private RidgeDetector ridgeDetector;

		/// <summary>Initialises a new instance of the <see cref="AbstractedLinesFilter"/> class.
		public AbstractedLinesFilter()
		{
			// default parameters (tone renderer)
			LightDirection = new Vector3(0.2f, 0.2f, -1.0f);
			IsPosition = false;
			Ambient    = 0;
			Diffuse    = 1;
			Specular   = 1;
			Shininess  = 2;
			ToonStep   = new Vector4(0.7f, 0.3f, 0.4f, 0.6f);

			// default parameter (tone blur)
			Sigma      = 8;

			// default parameters (ridge detector)
			LineWidth  = 8;
			NumSteps   = 5;
			Beta       = 0.3f;

			RidgesD = new Vector3(0.01f, 0.05f, 1.0f);
			RidgesS = new Vector3(0.01f, 0.05f, 1.0f);
			Valleys = new Vector3(0.01f, 0.05f, 1.0f);

			//BlendMode   = BlendMode.Additive;
			//BlendFactor = 0.5f;
			//Saturation  = 3.0f;
			BlendMode   = BlendMode.Multicative;
			BlendFactor = 0.6f;
			Saturation  = 1.4f;
	
			// create nodes outside the current topology (parameters are set later)
			toneRenderer  = new ToneRendererNode() { Topology = null, ToneOutput = ToneOutput.Mixed };
			toneBlur      = new ToneBlurNode()     { Topology = null };
			ridgeDetector = new RidgeDetector()    { Topology = null };

			AddInputPins("colour", "points");
			AddOutputPin("out");
		}

		/// <summary>Initialisation code, executed just before the first StatusUpdate.</summary>
		public override void Init()
		{
			toneRenderer.Init();
			toneBlur.Init();
			ridgeDetector.Init();
		}

		public override void Process()
		{
			// get inputs
			var frames = GetInputFrames();
			if (PassOnEndOfStreamFrame(frames)) return;
			var textures = CastFramesTo<TextureFrame>(frames);
			AssertFrameProperties(textures, FrameProps.Size);

			// set parameters
			toneRenderer.LightDirection = LightDirection;
			toneRenderer.IsPosition     = IsPosition;
			toneRenderer.Ambient        = Ambient;
			toneRenderer.Diffuse        = Diffuse;
			toneRenderer.Specular       = Specular;
			toneRenderer.Shininess      = Shininess;
			toneRenderer.ToonStep       = ToonStep;

			toneBlur.Sigma              = Sigma;

			ridgeDetector.LineWidth     = LineWidth;
			ridgeDetector.NumSteps      = NumSteps;
			ridgeDetector.Beta          = Beta;
			ridgeDetector.RidgesDLower  = RidgesD.X;
			ridgeDetector.RidgesDUpper  = RidgesD.Y;
			ridgeDetector.RidgesDAlpha  = RidgesD.Z;
			ridgeDetector.RidgesSLower  = RidgesS.X;
			ridgeDetector.RidgesSUpper  = RidgesS.Y;
			ridgeDetector.RidgesSAlpha  = RidgesS.Z;
			ridgeDetector.ValleysLower  = Valleys.X;
			ridgeDetector.ValleysUpper  = Valleys.Y;
			ridgeDetector.ValleysAlpha  = Valleys.Z;
			ridgeDetector.BlendMode     = BlendMode;
			ridgeDetector.BlendFactor   = BlendFactor;
			ridgeDetector.Saturation    = Saturation;

			var tone  = toneRenderer.Draw( new Dictionary<string, TextureFrame> { { "colour", textures["colour"] }, { "depth", textures["points"] } });
			var blur  = toneBlur.Draw(     new Dictionary<string, TextureFrame> { { "in", tone } });
			var ridge = ridgeDetector.Draw(new Dictionary<string, TextureFrame> { { "in", blur }, { "video", textures["colour"] } });

			PushFrame(ridge);
			tone.RefDec();
			blur.RefDec();
			ridge.RefDec();
			textures.RefDec();
		}
	}
}
