﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: ToneRenderer.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System.Collections.Generic;
using System.ComponentModel;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib.OpenGL;

namespace RGBZcamera.Library.AbstractedLines
{
	public enum ToneOutput
	{
		Mixed          =  0,
		Diffuse        =  1,
		Specular       =  2,
		Toon           =  3,
		Normals        =  4,
		VisibleNormals =  5,
		Alignment      =  6,
		HalfColour     =  7,
		ToonRelighting =  8,
		Video          =  9,
		Depth          = 10,
		Pointcloud     = 11,
		Distance       = 12,
	}

	public class ToneRendererNode : FboNode
	{
		#region Properties
//=================================================================================================

		[Uniform(), Category("Tone renderer"), DisplayName("Light direction"), Description("Position or direction of the single point light source.")]
		public Vector3 LightDirection { get; set; }

		[Uniform(), Category("Tone renderer"), DisplayName("Is position?"), Description("If set to true, will treat light position as a direction instead.")]
		public bool IsPosition { get; set; }

		[Uniform(), Category("Tone renderer"), Description("The amount of ambient illumination to use in the diffuse and toon shaded images.")]
		public float Ambient { get; set; }

		[Uniform(), Category("Tone renderer"), Description("The amount of diffuse illumination to use in the diffuse and toon shaded images.")]
		public float Diffuse { get; set; }

		[Uniform(), Category("Tone renderer"), Description("The amount of specular illumination to use in the specular and toon shaded images.")]
		public float Specular { get; set; }

		[Uniform(), Category("Tone renderer"), Description("The shininess of the material, the larger the more mirror-like.")]
		public float Shininess { get; set; }

		[Uniform(), Category("Tone renderer"), Description("The parameters of the toon step function: x + y * smoothstep(z, w, phong).")]
		public Vector4 ToonStep { get; set; }

		[Uniform(), Category("Tone renderer"), DefaultValue(ToneOutput.Mixed), DisplayName("Tone output"), Description("Select the output of this node.")]
		public ToneOutput ToneOutput { get; set; }

//=================================================================================================
		#endregion // Properties

		public ToneRendererNode()
		{
			// default settings
			LightDirection = new Vector3(0.2f, 0.2f, -1.0f);
			IsPosition = false;
			Ambient    = 0;
			Diffuse    = 1;
			Specular   = 1;
			Shininess  = 2;
			ToonStep   = new Vector4(0.7f, 0.3f, 0.4f, 0.6f);
			ToneOutput = ToneOutput.Mixed;

			AddInputPins("colour", "depth");
			AddOutputPin("out");
		}

		#region Rendering functions
//=================================================================================================

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.Mesh.vert", ShaderType.FragmentShader),
				Utils.CreateShaderFromResource("AbstractedLines.Shaders.ToneRenderer.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Name = "depth",  Unit = 0 },
				new TextureBinding { Name = "colour", Unit = 1 },
			};

			base.Init();
		}

		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			// allocate a texture that is identical to the input
			var renderTexture  = TextureFrame.CreateToMatch(inputs["colour"]);
			renderTexture.Wrap = TextureWrapMode.MirroredRepeat; // TODO: clamp to edge
			renderTexture.InternalFormat = PixelInternalFormat.Rgba32f;
			renderTexture.Alloc();

			Draw(inputs, renderTexture.Texture);

			return renderTexture;
		}

//=================================================================================================
		#endregion // Rendering functions
	}
}
