﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: FlowAverager.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib;
using VideoLib.OpenGL;
using OpenTK.Graphics.OpenGL;
using System.ComponentModel;

namespace RGBZcamera.Library
{
	/// <summary>Exponentially averages the incoming frames using avg' = w * in + (1 - w) * avg.</summary>
	public class FlowAverager : FboNode
	{
		/// <summary>Gets or sets the weight given to the current frame.</summary>
		[Uniform(), Description("Averaging weight given to the current frame")]
		public float Weight { get; set; }


		/// <summary>Initialises a new instance of the <see cref="ExponentialAverageFilter"/> class.</summary>
		public FlowAverager()
		{
			Weight = 0.2f;

			AddDefaultPins();
			AddInputPin("flow");
		}

		private TextureFrame previousAverage;

		/// <summary>Processes some frames, to produce new frames.</summary>
		public override void Process()
		{
			// get latest frame
			var frames = GetInputFrames();
			if (PassOnEndOfStreamFrame(frames)) return;
			var textures = CastFramesTo<TextureFrame>(frames);

			// draw quads
			var result = Draw(textures);
			PushFrame(result);

			// free previous frame
			if (previousAverage != null)
				previousAverage.RefDec();

			// keep the result for the next frame
			previousAverage = result;

			// clean up
			textures.RefDec();
		}

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.FlowAverager.frag")
			);

			textureMap = new List<TextureBinding>() {
				new TextureBinding { Pin = "in",   Uniform = "current",  Unit = 0 },
				new TextureBinding { Pin = null,   Uniform = "previous", Unit = 1 },
				new TextureBinding { Pin = "flow", Uniform = "flow",     Unit = 2 },
			};

			base.Init();
		}

		public override TextureFrame Draw(Dictionary<string,TextureFrame> inputs)
		{
			// allocate a texture that is identical to the input
			var resultTexture  = TextureFrame.CreateToMatch(inputs["in"]);
			resultTexture.Wrap = TextureWrapMode.ClampToEdge;
			resultTexture.Alloc();

			// bind texture of previous average
			if (previousAverage != null)
				previousAverage.Bind(TextureUnit.Texture1);

			// for first frame, use weight = 1 to ignore previous average (as it is undefined)
			float w = 0;
			if (previousAverage == null) { w = Weight; Weight = 1; }

			Draw(inputs, resultTexture.Texture);

			// return to actual weight for next frame
			if (previousAverage == null) Weight = w;

			return resultTexture;
		}
	}
}
