﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: DepthBlur.frag 537 2012-04-04 15:35:42Z cr333 $
*/
#version 120
#extension GL_EXT_gpu_shader4 : enable

// input tone images
uniform sampler2D input;

// standard deviation of the Gaussian blur
uniform float sigma;

// Applies a Gaussian blur to the first two channels of the input image (the diffuse and specular
// tone images), while leaving the third channel (toon shaded version) untouched.
void main(void)
{
	vec2 texelSize = 1.0 / textureSize2D(input, 0).xy;

	// colour of central pixel
	vec3 pix = texture2D(input, gl_FragCoord.yx * texelSize).rgb;
	
	// homogeneous accumulator (i.e. acc.z is the sum of weights)
	vec4 acc = vec4(pix.rgb, 1);
	
	// precompute common term
	float ss = -0.5f / (sigma * sigma);
	
	// set kernel size to two sigma
	int radius = int(ceil(2.0 * sigma));
	
	// apply horizontal blur while transposing the image
	// (this allows the same shader to be applied twice for a 2D Gaussian blur)
	for(int i = 1; i <= radius; i++)
	{
		// the two pixels at distance i from the centre share the same weight
		vec3 pix1 = texture2D(input, (gl_FragCoord.yx + vec2(-i, 0)) * texelSize).rgb;
		vec3 pix2 = texture2D(input, (gl_FragCoord.yx + vec2( i, 0)) * texelSize).rgb;
		acc += exp(i * i * ss) * vec4(pix1 + pix2, 2);
	}
	
	// output: unfiltered XYZ coordinates, blurred depth as 4th coordinate
	//gl_FragColor = vec4(pix.rgb, acc.z / acc.w);

	gl_FragColor = vec4(acc.rgb / acc.w, acc.z / acc.w);
}