﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: MeshTools.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using VideoLib.Frames;
using OpenTK.Graphics.OpenGL;
using OpenTK;
using System;

namespace RGBZcamera.Library
{
	/// <summary>Contains various tools relating to meshes.</summary>
	public static class MeshTools
	{
		/// <summary>Updates the given index buffer to represent a base mesh of the given size.</summary>
		/// <param name="width">The width of the base mesh.</param>
		/// <param name="height">The height of the base mesh.</param>
		/// <param name="indexBufferID">The index buffer ID.</param>
		/// <param name="numIndices">The number of indices of the base mesh.</param>
		public static unsafe void UpdateBaseMeshIndexBuffer(int width, int height, int indexBufferID, out int numIndices)
		{
			// read out size of buffer
			int bufferSize = 0;
			GL.BindBuffer(BufferTarget.ElementArrayBuffer, indexBufferID);
			GL.GetBufferParameter(BufferTarget.ElementArrayBuffer, BufferParameterName.BufferSize, out bufferSize);

			// has the buffer the right size? -> no update required
			numIndices = 3 * 2 * (width - 1) * (height - 1);
			if (bufferSize == numIndices * sizeof(uint))
			{
				GL.BindBuffer(BufferTarget.ElementArrayBuffer, 0);
				return;
			}

			// else: reallocate buffer if too small
			if (bufferSize < numIndices * sizeof(uint))
				GL.BufferData(BufferTarget.ElementArrayBuffer, (IntPtr)(numIndices * sizeof(uint)), IntPtr.Zero, BufferUsageHint.StaticDraw);

			// map the index buffer, so we can write directly into it
			var indexBuffer = GL.MapBuffer(BufferTarget.ElementArrayBuffer, BufferAccess.WriteOnly);
			uint* Indices = (uint*)indexBuffer.ToPointer();

			int index = 0;
			for (int y = 0; y < height - 1; y++)
			{
				for (int x = 0; x < width - 1; x++)
				{
					// vertex indices
					uint TL = (uint)((y + 0) * width + (x + 0)); // TL--TR   top-left/right
					uint TR = (uint)((y + 0) * width + (x + 1)); // | \  |
					uint BL = (uint)((y + 1) * width + (x + 0)); // |  \ |
					uint BR = (uint)((y + 1) * width + (x + 1)); // BL--BR   bottom-left/right

					// first triangle (top-right)
					Indices[index++] = TL;
					Indices[index++] = TR;
					Indices[index++] = BR;

					// second triangle (bottom-left)
					Indices[index++] = TL;
					Indices[index++] = BR;
					Indices[index++] = BL;
				}
			}

			GL.UnmapBuffer(BufferTarget.ElementArrayBuffer);
			GL.BindBuffer(BufferTarget.ElementArrayBuffer, 0);
		}

		/// <summary>Updates the given vertex buffer to contain a base mesh of the given size.</summary>
		/// <remarks>Uses uints as indices.</remarks>
		/// <param name="width">The width of the base mesh.</param>
		/// <param name="height">The height of the base mesh.</param>
		/// <param name="vertexBufferID">The vertex buffer ID.</param>
		/// <param name="numVertices">The number of vertices in the base mesh.</param>
		public static unsafe void UpdateBaseMeshVertexBuffer(int width, int height, int vertexBufferID, out int numVertices)
		{
			// read out size of buffer
			int bufferSize = 0;
			GL.BindBuffer(BufferTarget.ArrayBuffer, vertexBufferID);
			GL.GetBufferParameter(BufferTarget.ArrayBuffer, BufferParameterName.BufferSize, out bufferSize);

			// has the buffer the right size? -> no update required
			numVertices = width * height;
			if (bufferSize == numVertices * Vector3.SizeInBytes)
			{
				GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
				return;
			}

			// else: reallocate buffer if too small
			if (bufferSize < numVertices * Vector3.SizeInBytes)
				GL.BufferData(BufferTarget.ArrayBuffer, (IntPtr)(numVertices * Vector3.SizeInBytes), IntPtr.Zero, BufferUsageHint.StaticDraw);

			// map the vertex buffer, so we can write directly into it
			var vertexBuffer  = GL.MapBuffer(BufferTarget.ArrayBuffer, BufferAccess.WriteOnly);
			Vector3* Vertices = (Vector3*)vertexBuffer.ToPointer();

			int index = 0;
			for (int y = 0; y < height; y++)
			{
				for (int x = 0; x < width; x++)
				{
					Vertices[index++] = new Vector3(x, y, 0);
					//Vertices[index++] = new Vector3(x, y, 10);
					//Vertices[index++] = new Vector3(x / (float)width, y / (float)height, 0);
					//Vertices[index++] = new Vector3((x + 0.5f) / width, (y + 0.5f) / height, 0); // [0..1] x [0..1]
				}
			}

			GL.UnmapBuffer(BufferTarget.ArrayBuffer);
			GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
		}

		/// <summary>Computes and uploads the vertex normals for a given point cloud.</summary>
		/// <param name="bitmap">The new point cloud.</param>
		[Obsolete]
		public static unsafe void UpdateNormalBuffer(BitmapFrame bitmap, int normalBufferID)
		{
			// check if the buffer is large enough and reallocate if not
			int bufferSize = 0, numVertices = bitmap.Width * bitmap.Height;
			GL.BindBuffer(BufferTarget.ArrayBuffer, normalBufferID);
			GL.GetBufferParameter(BufferTarget.ArrayBuffer, BufferParameterName.BufferSize, out bufferSize);
			if (bufferSize < numVertices * Vector3.SizeInBytes)
				GL.BufferData(BufferTarget.ArrayBuffer, (IntPtr)(numVertices * Vector3.SizeInBytes), IntPtr.Zero, BufferUsageHint.DynamicDraw);

			// map the normal buffer, so we can write directly into it
			var normalBuffer = GL.MapBuffer(BufferTarget.ArrayBuffer, BufferAccess.WriteOnly);

			fixed (byte* bytes = bitmap.Data)
			{
				Vector3* vertices = (Vector3*)bytes;
				Vector3* vertexNormals = (Vector3*)normalBuffer.ToPointer();

				for (int y = 0; y < bitmap.Height; y++)
				{
					for (int x = 0; x < bitmap.Width; x++)
					{
						var offset = y * bitmap.Width + x;

						// Step 1. look up point coordinates ------------------------------------------

						// this point (centre)
						Vector3 pointC = vertices[offset];
						Vector3 pointT, pointL, pointR, pointB; // points to the top, left, right and bottom

						// top point defined
						if (y - 1 >= 0) pointT = vertices[offset - bitmap.Width];
						else pointT = Vector3.Zero;

						// left point defined
						if (x - 1 >= 0) pointL = vertices[offset - 1];
						else pointL = Vector3.Zero;

						// right point defined
						if (x + 1 < bitmap.Width) pointR = vertices[offset + 1];
						else pointR = Vector3.Zero;

						// bottom point defined
						if (y + 1 < bitmap.Height) pointB = vertices[offset + bitmap.Width];
						else pointB = Vector3.Zero;

						// Step 2. compute vectors to the 4 neighbouring points -----------------------

						Vector3 vectorT, vectorL, vectorR, vectorB;
						Vector3.Subtract(ref pointT, ref pointC, out vectorT);
						Vector3.Subtract(ref pointL, ref pointC, out vectorL);
						Vector3.Subtract(ref pointR, ref pointC, out vectorR);
						Vector3.Subtract(ref pointB, ref pointC, out vectorB);

						// Step 3. compute face normals -----------------------------------------------

						// Layout of triangles for normal computation
						// (note: this differs from the mesh layout)
						//
						//       T
						//      /|\
						//     / | \
						//    /  |  \
						//   L---C---R
						//    \  |  /
						//     \ | /
						//      \|/
						//       B

						Vector3 normal = Vector3.Zero;
						int normals = 0;

						// top-left triangle
						if (pointT != Vector3.Zero && pointL != Vector3.Zero)
						{
							Vector3 faceNormal;
							Vector3.Cross(ref vectorT, ref vectorL, out faceNormal);
							faceNormal.NormalizeFast();
							Vector3.Add(ref normal, ref faceNormal, out normal);
							normals++;
						}

						// top-right triangle
						if (pointT != Vector3.Zero && pointR != Vector3.Zero)
						{
							Vector3 faceNormal;
							Vector3.Cross(ref vectorR, ref vectorT, out faceNormal);
							faceNormal.NormalizeFast();
							Vector3.Add(ref normal, ref faceNormal, out normal);
							normals++;
						}

						// bottom-left triangle
						if (pointB != Vector3.Zero && pointL != Vector3.Zero)
						{
							Vector3 faceNormal;
							Vector3.Cross(ref vectorB, ref vectorR, out faceNormal);
							faceNormal.NormalizeFast();
							Vector3.Add(ref normal, ref faceNormal, out normal);
							normals++;
						}

						// bottom-right triangle
						if (pointB != Vector3.Zero && pointR != Vector3.Zero)
						{
							Vector3 faceNormal;
							Vector3.Cross(ref vectorL, ref vectorB, out faceNormal);
							faceNormal.NormalizeFast();
							Vector3.Add(ref normal, ref faceNormal, out normal);
							normals++;
						}

						// Step 4. average face normals to get vertex normals -------------------------

						Vector3.Divide(ref normal, normals, out normal);
						normal.NormalizeFast();
						vertexNormals[y * bitmap.Width + x] = normal;
					}
				}
			}

			GL.UnmapBuffer(BufferTarget.ArrayBuffer);
		}
	}
}