﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: Mesh.vert 537 2012-04-04 15:35:42Z cr333 $
*/
#version 130
#extension GL_EXT_gpu_shader4 : enable

// returns the coordinates of the current vertex
vec3 getVertex(sampler2D coordinates, vec2 position)
{
	// texel coordinate from texture coordinate
	ivec2 texelpos = ivec2(position + 0.5);

	// clamp texel coordinate to valid range
	texelpos = clamp(texelpos, ivec2(0), textureSize(coordinates, 0) - ivec2(1));

	// fetch vertex coordinates, in world coordinates
	return texelFetch(coordinates, texelpos, 0).xyz;
}

// Computes the vertice normal by averaging face normals.
//
// Notes:
//   - coordinates of the vertex are passed in to save a texture read
//   - the returned normal is *not* normalised
vec3 getVertexNormal(sampler2D coordinates, vec2 position, vec3 vertexCoords)
{
	// look up coordinates of the 4 neighbours, in world coordinates
	vec3 pc = vertexCoords;
	vec3 pt = texelFetch(coordinates, ivec2(position + 0.5) + ivec2( 0, -1), 0).xyz;
	vec3 pl = texelFetch(coordinates, ivec2(position + 0.5) + ivec2(-1,  0), 0).xyz;
	vec3 pr = texelFetch(coordinates, ivec2(position + 0.5) + ivec2( 1,  0), 0).xyz;
	vec3 pb = texelFetch(coordinates, ivec2(position + 0.5) + ivec2( 0,  1), 0).xyz;

	// Compute vertex normal by averaging all face normals.
	//
	// Layout of triangles for normal computation
	// (note: this differs from the mesh layout)
	//
	//       T
	//      /|\
	//     / | \
	//    /  |  \
	//   L---C---R
	//    \  |  /
	//     \ | /
	//      \|/
	//       B

	//// ignore face normals if two points are identical (e.g. on the border of the mesh, due to texture clamping)
	//// the face normals contribute proportionally to the area of their triangles
	//vec4 normal = vec4(0.0);
	//if(distance(pt, pc) > 0 && distance(pl, pc) > 0) { normal += vec4(cross(pt - pc, pl - pc), 1); }
	//if(distance(pl, pc) > 0 && distance(pb, pc) > 0) { normal += vec4(cross(pl - pc, pb - pc), 1); }
	//if(distance(pb, pc) > 0 && distance(pr, pc) > 0) { normal += vec4(cross(pb - pc, pr - pc), 1); }
	//if(distance(pr, pc) > 0 && distance(pt, pc) > 0) { normal += vec4(cross(pr - pc, pt - pc), 1); }
	//return normal.xyz / normal.w;
	
	//// the face normals contribute proportionally to the area of their triangles
	//return (cross(pt - pc, pl - pc) + cross(pl - pc, pb - pc) + cross(pb - pc, pr - pc) + cross(pr - pc, pt - pc)) / 4;
	
	//// To deal with pixels near depth discontinuities, we need to check which of the neighbouring pixels
	//// are outliers, i.e. are further away (in depth) from the central pixel than the pixel opposite.
	//// For example, the right pixel is considered invalid if its depth distance to the central pixel
	//// is much larger than the one of the left pixel.
	//float threshold = 25;
	//bool vt = abs(pt.z - pc.z) <= threshold * abs(pb.z - pc.z);
	//bool vl = abs(pl.z - pc.z) <= threshold * abs(pr.z - pc.z);
	//bool vb = abs(pb.z - pc.z) <= threshold * abs(pt.z - pc.z);
	//bool vr = abs(pr.z - pc.z) <= threshold * abs(pl.z - pc.z);
	////if(!vr || !vl || !vb || !vt) return vec3(0, 0, -1); // highlight invalid pixels
	//
	//// average the valid face normals proportionally to the area of their triangles
	//vec4 normal = vec4(0.0);
	//if(vt && vl) normal += vec4(cross(pt - pc, pl - pc), 1);
	//if(vl && vb) normal += vec4(cross(pl - pc, pb - pc), 1);
	//if(vb && vr) normal += vec4(cross(pb - pc, pr - pc), 1);
	//if(vr && vt) normal += vec4(cross(pr - pc, pt - pc), 1);
	//return normal.xyz / normal.w;
	
	// Average all face normals that are close to the most frontal one (largest z component)

	// unit face normals
	vec3 ntl = normalize(cross(pt - pc, pl - pc));
	vec3 nbl = normalize(cross(pl - pc, pb - pc));
	vec3 nbr = normalize(cross(pb - pc, pr - pc));
	vec3 ntr = normalize(cross(pr - pc, pt - pc));
	
	// reference normal: select most frontal one
	vec3 ref = ntl;
	if(abs(ntr.z) > abs(ref.z)) ref = ntr;
	if(abs(nbl.z) > abs(ref.z)) ref = nbl;
	if(abs(nbr.z) > abs(ref.z)) ref = nbr;
	
	// average face normals that are within 30 degrees of the reference normal
	vec4 normal = vec4(0.0);
	float threshold = cos(radians(30.0));
	if(dot(ref, ntl) >= threshold) normal += vec4(ntl, 1);
	if(dot(ref, ntr) >= threshold) normal += vec4(ntr, 1);
	if(dot(ref, nbl) >= threshold) normal += vec4(nbl, 1);
	if(dot(ref, nbr) >= threshold) normal += vec4(nbr, 1);
	return normal.xyz / normal.w;
}