﻿// $Id: LibAvSink.cs 562 2012-04-23 09:10:58Z cr333 $
using System;
using System.Collections.Generic;
using System.Text;
using VideoLib.Frames;
using System.Runtime.InteropServices;
using System.IO;
using LibAvWrap;

namespace VideoLib.LibAv
{
	/// <summary>Writes a stream of BitmapFrames to a video file.</summary>
	public class LibAvSink : Node, IDisposable
	{
		///<summary>The filename or URL of the underlying video.</summary>
		public string Filename { get; set; }

		///<summary>Frame rate of the video.</summary>
		public int FrameRate { get; set; }

		///<summary>Target bit rate of the video.</summary>
		public int BitRate { get; set; }

		///<summary>Size of the buffer used when writing the video.</summary>
		public int VideoBufferSize { get; set; }
		
		/// <summary>Width of video frames. Set by first frame.</summary>
		public int Width { get; protected set; }

		/// <summary>Height of video frames. Set by first frame.</summary>
		public int Height { get; protected set; }

		/// <summary>Human-friendly node name.</summary>
		public override string Name
		{
			get { return String.Format("{0}({1})", base.Name, Path.GetFileName(Filename)); }
		}
		
		LibAvFormatContext formatContext = null;
		LibAvStream videoStream = null;
		IntPtr videoBuffer = IntPtr.Zero;

		LibAvFrame picture = null;
		LibAvFrame tmp_picture = null;

		int frame_count = 0;
		bool open = false;

		/// <summary>Initialises the underlying libav.</summary>
		static LibAvSink()
		{
			// Initialize libavformat and register all the muxers, demuxers and protocols.
			LibAvUtils.RegisterAll();
			LibAvUtils.UsePrintfLogger();
		}

		/// <summary>Constructs a new video sink, to write to the given video file.</summary>
		/// <param name="filename">Filename of the video to write.</param>
		public LibAvSink(string filename) : base()
		{
			Filename  = filename;
			FrameRate = 25;
			BitRate = 100000000; // virtually uncompressed (100 mbps)

			VideoBufferSize = 200000;
			AddInputPin("in");
		}

		/// <summary>Processes a BitmapFrame.</summary>
		public override void Process()
		{
			Frame frame = GetInputPin("in").GetFrame();

			// the first frame of the stream defines the video size
			if(!open)
			{
				Width  = (frame as VideoFrame).Width;
				Height = (frame as VideoFrame).Height;

				// open the video file and all streams for writing
				OpenVideoFile();
			}

			if (frame is BitmapFrame)
			{
				WriteVideoFrame((BitmapFrame)frame);
			}

			// close video streams and file at the end of the video stream
			if (frame is EndOfStreamFrame)
			{
				CloseVideoFile();
				Status = NodeStatus.Finished;
			}
		}

		/// <summary>Closes the video file and cleans up.</summary>
		protected override void Dispose(bool disposing)
		{
			CloseVideoFile();
			base.Dispose(disposing);
		}

		private void OpenVideoFile()
		{
			// Auto-detect the output format from the name
			LibAvOutputFormat outputFormat = LibAvUtils.GuessType("", Filename, "");

			// allocate the output media context
			formatContext = new LibAvFormatContext() { Filename = Filename, OutputFormat = outputFormat };

			// add the audio and video streams using the default format codecs and initialize the codecs
			if (formatContext.OutputFormat.VideoCodec != CodecID.NONE)
				videoStream = AddVideoStream();

			// set the output parameters (must be done even if no parameters).
			if (formatContext.SetParameters(IntPtr.Zero) < 0)
				throw new Exception("Invalid output format parameters.");

			#if DEBUG
				// print stream info for debug
				formatContext.DumpFormat(0, Filename, true);
			#endif

			// now that all the parameters are set, we can open the audio and
			// video codecs and allocate the necessary encode buffers
			if (videoStream.IntPointer != IntPtr.Zero)
				OpenVideo();

			// open the output file, if needed
			if ((formatContext.OutputFormat.Flags & Av.AvFormatNoFile) == 0)
			{
				MakeDirs(Filename);
				LibAvUtils.ThrowIfError(LibAvUtils.UrlFOpen(formatContext, Filename, Av.UrlWriteOnly));
			}

			// write the stream header, if any
			formatContext.WriteHeader();

			open = true;
		}

		/// <summary>Makes sure that the parent directory exists.</summary>
		/// <param name="filename">Some filename.</param>
		private static void MakeDirs(string filename)
		{
			var parentDir = new FileInfo(filename).Directory;
			if (!parentDir.Exists)
				parentDir.Create();
		}

		private void CloseVideoFile()
		{
			if (open)
			{
				open = false;

				// write the trailer, if any
				formatContext.WriteTrailer();

				// close each codec
				videoStream.Codec.Close();
				if (videoBuffer != IntPtr.Zero)
				{
					Marshal.FreeHGlobal(videoBuffer);
					videoBuffer = IntPtr.Zero;
				}

				//// free the streams
				//for(i = 0; i < formatContext->nb_streams; i++) {
				//    av_freep(&formatContext->streams[i]->codec);
				//    av_freep(&formatContext->streams[i]);
				//}

				// close the output file
				if ((formatContext.OutputFormat.Flags & Av.AvFormatNoFile) == 0)
					LibAvUtils.UrlFClose(formatContext.GetPb());

				// free the stream
				formatContext.Dispose();
			}
		}

		void WriteVideoFrame(BitmapFrame bitmapFrame)
		{
			LibAvWrap.PixelFormat pixelFormat = LibAvWrap.PixelFormat.None;
			switch (bitmapFrame.Format)
			{
				case VideoLib.Frames.PixelFormat.BGR24: pixelFormat = LibAvWrap.PixelFormat.BGR24; break;
				case VideoLib.Frames.PixelFormat.BGR32: pixelFormat = LibAvWrap.PixelFormat.BGR32; break;
			}

			LibAvFrame tempPicture = new LibAvFrame(bitmapFrame.Data, pixelFormat, bitmapFrame.Width, bitmapFrame.Stride, bitmapFrame.Height);
			WriteVideoFrame(tempPicture, pixelFormat);
			tempPicture.Dispose();
		}

		void WriteVideoFrame(LibAvFrame pict, LibAvWrap.PixelFormat pixelFormat)
		{
			LibAvCodecContext c = videoStream.Codec;

			// Convert the frame if it does not have the right format
			if (c.PixFmt != pixelFormat)
			{
				//LibAvFrame.Convert(pict, pixelFormat, picture, c.PixFmt, c.Width, c.Height);
				pict.ConvertTo(picture);
			}
			else
			{
				picture = pict;
			}

			// encode the image
			int out_size = c.EncodeVideo(videoBuffer, VideoBufferSize, picture);

			// if zero size, it means the image was buffered
			if (out_size > 0)
			{
				LibAvPacket pkt = new LibAvPacket();
				c.SetPacketPts(pkt, videoStream);
				c.HandlePacketFlagKey(pkt);

				pkt.StreamIndex = videoStream.Index;
				pkt.SetData(videoBuffer);
				pkt.Size = out_size;

				// write the compressed frame in the media file
				LibAvUtils.ThrowIfError(formatContext.WriteFrame(pkt));
				pkt.Free();
			}

			frame_count++;
		}

		// add a video output stream
		private LibAvStream AddVideoStream()
		{
			LibAvStream st = formatContext.CreateNewStream(0);

			LibAvCodecContext c = st.Codec;
			c.CodecId = formatContext.OutputFormat.VideoCodec;
			c.MediaType = MediaType.Video;

			// put sample parameters
			c.BitRate = BitRate;

			// resolution must be a multiple of two -- TODO
			c.Width = Width;
			c.Height = Height;

			// time base: this is the fundamental unit of time (in seconds) in terms
			//   of which frame timestamps are represented. for fixed-fps content,
			//   timebase should be 1/framerate and timestamp increments should be
			//   identically 1.
			c.SetTimeBase(1, FrameRate);
			c.GopSize = 12; // emit one intra frame every twelve frames at most
			c.PixFmt = LibAvWrap.PixelFormat.YUV420P; // TODO: make parameter

			// just for testing, we also add B frames
			if (c.CodecId == CodecID.MPEG2VIDEO)
				c.MaxBFrames = 2;

			// Needed to avoid using macroblocks in which some coeffs overflow. This does not happen
			// with normal video, it just happens here as the motion of the chroma plane does not match the luma plane.
			if (c.CodecId == CodecID.MPEG1VIDEO)
				c.MbDecision = 2;

			//// some formats want stream headers to be separate
			//if (formatContext.OutputFormat.Name == "mp4" || formatContext.OutputFormat.Name == "mov" || formatContext.OutputFormat.Name == "3gp")
			//    c.Flags |= CodecFlags.GlobalHeader;

			return st;
		}

		private void OpenVideo()
		{
			// open the video encoder
			LibAvCodecContext c = videoStream.Codec;
			LibAvUtils.ThrowIfError(c.OpenEncoder());

			// allocate output buffer
			if ((formatContext.OutputFormat.Flags & Av.AvRawPicture) == 0)
				videoBuffer = Marshal.AllocHGlobal(VideoBufferSize);

			// allocate the encoded raw picture
			picture = new LibAvFrame(c.PixFmt, c.Width, c.Height);
			// TODO: error handling

			// if the output format is not YUV420P, then a temporary YUV420P picture is needed too.
			// It is then converted to the required output format
			if (c.PixFmt != LibAvWrap.PixelFormat.YUV420P)
			{
				tmp_picture = new LibAvFrame(LibAvWrap.PixelFormat.YUV420P, c.Width, c.Height);
				// TODO: error handling
			}
		}
	}
}
