﻿// $Id: StereoVideoControl.cs 557 2012-04-18 12:50:25Z cr333 $
using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib.Frames;
using VideoLib.OpenGL;
using VideoLib.Sinks;

namespace VideoLib.OpenGL
{
	/// <summary>Enumerates the stereo display modes.</summary>
	/// <remarks>The numbering may seem a bit odd, but it results in a reasonable order
	/// when all values are listed, for example in a PropertyGrid.</remarks>
	[Flags]
	public enum StereoMode
	{
	//---- basic types ----------------------------------------------------------------------------

		/// <summary>Show only the left view.</summary>
		LeftOnly = 1 << 1,

		/// <summary>Show only the right view.</summary>
		RightOnly = 1 << 2,

		/// <summary>Swap the left and right views.</summary>
		SwapViews = 1 << 3,

		/// <summary>Blend both views together evenly.</summary>
		Overlay = 1 << 4,

		/// <summary>Show row-interlaced stereo image with the left view in even lines.</summary>
		RowInterlacedEven = 1 << 5,

		/// <summary>Greyscale anaglyph mode. Also need to specify per-view colour components.</summary>
		GreyAnaglyph = 1 << 20,

		/// <summary>Full-colour anaglyph mode. Also need to specify per-view colour components.</summary>
		FullColourAnaglyph = 1 << 21,

		/// <summary>Show row-interlaced stereo image with the left view in odd lines.</summary>
		RowInterlacedOdd = RowInterlacedEven | SwapViews,

	//---- anaglyph modes -------------------------------------------------------------------------

		/// <summary>Anaglyph mode with the red channel taken from the left image.</summary>
		LeftRed     = 1 << 13,

		/// <summary>Anaglyph mode with the green channel taken from the left image.</summary>
		LeftGreen   = 1 << 14,

		/// <summary>Anaglyph mode with the blue channel taken from the left image.</summary>
		LeftBlue    = 1 << 15,

		/// <summary>Anaglyph mode with the green and blue channels taken from the left image.</summary>
		LeftCyan    = LeftGreen | LeftBlue,

		/// <summary>Anaglyph mode with the red and green channels taken from the left image.</summary>
		LeftYellow  = LeftRed   | LeftGreen,

		/// <summary>Anaglyph mode with the red and blue channels taken from the left image.</summary>
		LeftMagenta = LeftRed   | LeftBlue,


		/// <summary>Anaglyph mode with the red channel taken from the right image.</summary>
		RightRed     = 1 <<  10,

		/// <summary>Anaglyph mode with the green channel taken from the right image.</summary>
		RightGreen   = 1 <<  11,

		/// <summary>Anaglyph mode with the blue channel taken from the right image.</summary>
		RightBlue    = 1 <<  12,

		/// <summary>Anaglyph mode with the green and blue channels taken from the right image.</summary>
		RightCyan    = RightGreen | RightBlue,

		/// <summary>Anaglyph mode with the red and green channels taken from the right image.</summary>
		RightYellow  = RightRed   | RightGreen,

		/// <summary>Anaglyph mode with the red and blue channels taken from the right image.</summary>
		RightMagenta = RightRed   | RightBlue,


		/// <summary>Full-colour red-blue anaglyph.</summary>
		FullColourAnaglyphRedBlue = FullColourAnaglyph | LeftRed | RightBlue,

		/// <summary>Full-colour red-cyan anaglyph.</summary>
		FullColourAnaglyphRedCyan = FullColourAnaglyph | LeftRed | RightCyan,

		/// <summary>Full-colour red-green anaglyph.</summary>
		FullColourAnaglyphRedGreen = FullColourAnaglyph | LeftRed | RightGreen,

		/// <summary>Full-colour green-red anaglyph.</summary>
		FullColourAnaglyphGreenRed = FullColourAnaglyph | LeftGreen | RightRed,

		/// <summary>Full-colour green-magenta anaglyph.</summary>
		FullColourAnaglyphGreenMagenta = FullColourAnaglyph | LeftGreen | RightMagenta,

		/// <summary>Full-colour blue-red anaglyph.</summary>
		FullColourAnaglyphBlueRed = FullColourAnaglyph | LeftBlue | RightRed,

		/// <summary>Full-colour cyan-red anaglyph.</summary>
		FullColourAnaglyphCyanRed = FullColourAnaglyph | LeftCyan | RightRed,

		/// <summary>Full-colour magenta-green anaglyph.</summary>
		FullColourAnaglyphMagentaGreen = FullColourAnaglyph | LeftMagenta | RightGreen,


		/// <summary>Grey red-blue anaglyph.</summary>
		GreyAnaglyphRedBlue = GreyAnaglyph | LeftRed | RightBlue,

		/// <summary>Grey red-cyan anaglyph.</summary>
		GreyAnaglyphRedCyan = GreyAnaglyph | LeftRed | RightCyan,

		/// <summary>Grey red-green anaglyph.</summary>
		GreyAnaglyphRedGreen = GreyAnaglyph | LeftRed | RightGreen,

		/// <summary>Grey green-red anaglyph.</summary>
		GreyAnaglyphGreenRed = GreyAnaglyph | LeftGreen | RightRed,

		/// <summary>Grey green-magenta anaglyph.</summary>
		GreyAnaglyphGreenMagenta = GreyAnaglyph | LeftGreen | RightMagenta,

		/// <summary>Grey blue-red anaglyph.</summary>
		GreyAnaglyphBlueRed = GreyAnaglyph | LeftBlue | RightRed,

		/// <summary>Grey cyan-red anaglyph.</summary>
		GreyAnaglyphCyanRed = GreyAnaglyph | LeftCyan | RightRed,

		/// <summary>Grey magenta-green anaglyph.</summary>
		GreyAnaglyphMagentaGreen = GreyAnaglyph | LeftMagenta | RightGreen,
	}

	/// <summary>Implements an OpenGL-based stereo video rendering control.</summary>
	public partial class StereoVideoControl : GLControl, IStereoVideoControl
	{
		/// <summary>Indicates if the control is currently displayed in the designer.</summary>
		protected bool InDesignMode
		{
			get { if (DesignMode) return true; return this.ParentForm.Site != null; }
		}

		/// <summary>Constructs an OpenGL-based video rendering control.</summary>
		public StereoVideoControl()
		{
			InitializeComponent();

			// set custom style
			this.SetStyle(ControlStyles.Selectable, false);
			this.SetStyle(ControlStyles.ResizeRedraw, true);

			// set default property values
			StereoMode          = StereoMode.FullColourAnaglyphRedCyan;
			PreserveAspectRatio = true;
			LinearInterpolation = true;
			ShowCaption         = false;
			Zoom                = 1.0f;

#if DEBUG
			Background = Color.Black;
#else
			Background = Color.FromKnownColor(KnownColor.Control);
#endif
		}

//=================================================================================================
		#region Fields and properties

		/// <summary>Gets or sets whether the aspect ratio of the input is preserved (false means stretch to fill). Default: true.</summary>
		[DefaultValue(true), Description("Indicates whether the aspect ratio of the input is preserved.")]
		public bool PreserveAspectRatio { get; set; }

		/// <summary>Gets or sets the zoom level (only used if AutoSize is true).</summary>
		[DefaultValue(1.0f), Description("The zoom level applied to the display.")]
		public float Zoom { get; set; }

		/// <summary>Gets or sets the background colour behind and surrounding the video.</summary>
		[Description("The background colour behind and surrounding the video.")]
		public Color Background { get; set; }

		/// <summary>Gets or sets whether to show the caption and FPS counter.</summary>
		[DefaultValue(false), Description("Indicates whether to show the caption and a framerate counter.")]
		public bool ShowCaption { get; set; }

		/// <summary>Gets or sets whether linear interpolation is enabled.</summary>
		[DefaultValue(true), Description("Enables linear interpolation of the displayed texture.")]
		public bool LinearInterpolation { get; set; }

		/// <summary>Gets or sets the stereo mode of the display.</summary>
		[DefaultValue(StereoMode.FullColourAnaglyphRedCyan), Description("The stereo mode to use.")]
		public StereoMode StereoMode { get { return stereoMode; } set { stereoMode = value; dirty = true; } }

		/// <summary>The texture that will hold the left video frames.</summary>
		private Texture textureLeft;

		/// <summary>The texture that will hold the right video frames.</summary>
		private Texture textureRight;

		/// <summary>Copy of the latest caption.</summary>
		private string caption = "";

		/// <summary>The stereo mode to use.</summary>
		private StereoMode stereoMode;

		#endregion // Fields and properties
//=================================================================================================
		#region Rendering functions

		private int programID;
		private bool dirty = true;

		/// <summary>Initialises the OpenGL state for this control.</summary>
		private void InitGL()
		{
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.StereoVideoControl.frag")
			);
		}

		/// <summary>Accepts a new pair of frames and a caption to be drawn onto this control.</summary>
		/// <param name="bitmapFrameL">The new left frame.</param>
		/// <param name="bitmapFrameR">The new right frame.</param>
		/// <param name="caption">The new caption (ignored for now).</param>
		public void Draw(BitmapFrame bitmapFrameL, BitmapFrame bitmapFrameR, string caption)
		{
			// only go ahead if it is safe to do so
			if (!Disposing && Created)
			{
				// need to draw the image on the UI thread
				BeginInvoke(new Action(delegate() { DrawImageInternal(bitmapFrameL, bitmapFrameR, caption); }));
			}
		}

		/// <summary>Accepts a new pair of frames and a caption to be drawn onto this control.</summary>
		/// <remarks>Needs to run on the GUI/OpenGL thread.</remarks>
		/// <param name="bitmapL">The new left image.</param>
		/// <param name="bitmapR">The new right image.</param>
		/// <param name="caption">The new caption (ignored for now).</param>
		private void DrawImageInternal(BitmapFrame bitmapL, BitmapFrame bitmapR, string caption)
		{
			// this handles most crashes when closing the form
			if (Disposing || !Created)
				return;

			this.caption = caption;

			// implement auto-sizing
			if (AutoSize)
			{
				Size bitmapSize = new Size((int)(Zoom * bitmapL.Width + 0.5f), (int)(Zoom * bitmapL.Height + 0.5f));
				this.Size   = bitmapSize;
				MinimumSize = bitmapSize;
				MaximumSize = bitmapSize;
			}

			MakeCurrent();

			if (textureLeft == null)
			{
				textureLeft = new Texture(bitmapL)
				{
					Filter = (LinearInterpolation ? TextureMagFilter.Linear : TextureMagFilter.Nearest),
					Wrap = TextureWrapMode.ClampToEdge // avoid fade-to-black borders in low-res images
				};
			}

			if (textureRight == null)
			{
				textureRight = new Texture(bitmapR)
				{
					Filter = (LinearInterpolation ? TextureMagFilter.Linear : TextureMagFilter.Nearest),
					Wrap = TextureWrapMode.ClampToEdge // avoid fade-to-black borders in low-res images
				};
			}

			// update the textures with new frame data
			textureLeft.Update(bitmapL);
			textureRight.Update(bitmapR);

			// ask for a full redraw
			Invalidate();
		}

		/// <summary>Renders the current video frame.</summary>
		private void Render()
		{
			// Set up matrices and viewport.
			GL.MatrixMode(MatrixMode.Projection); GL.LoadIdentity();
			GL.MatrixMode(MatrixMode.Modelview); GL.LoadIdentity();
			GL.Viewport(0, 0, Width, Height);

			// First clear the background, regardless of whether we have data.
			GL.DrawBuffer(DrawBufferMode.Back);
			GL.ClearColor(Background);
			GL.Clear(ClearBufferMask.ColorBufferBit);

			if (textureLeft != null && textureRight != null)
			{
				// The necessary setup for texturing without depth.
				GL.Enable(EnableCap.Texture2D);
				GL.Disable(EnableCap.DepthTest);
				GL.DepthMask(false);

				// Work out optimal size of texture to fit inside the control.
				float w, h;
				ComputeTargetSize(textureLeft, out w, out h);

				GL.UseProgram(programID);
				textureLeft.Bind(TextureUnit.Texture0);
				textureRight.Bind(TextureUnit.Texture1);

				var screenPos = PointToScreen(Point.Empty);
				GL.Uniform2(GL.GetUniformLocation(programID, "screenPos"), new Vector2(screenPos.X, screenPos.Y));
				if (dirty)
				{
					GL.Uniform1(GL.GetUniformLocation(programID, "inputL"), 0);
					GL.Uniform1(GL.GetUniformLocation(programID, "inputR"), 1);
					GL.Uniform1(GL.GetUniformLocation(programID, "stereoMode"), (int)StereoMode);
					dirty = false;
				}

				GL.Begin(BeginMode.Quads); // NB: texcords apply y-flip here
				GL.TexCoord2(0, 1); GL.Vertex2(-w, -h);
				GL.TexCoord2(1, 1); GL.Vertex2( w, -h);
				GL.TexCoord2(1, 0); GL.Vertex2( w,  h);
				GL.TexCoord2(0, 0); GL.Vertex2(-w,  h);
				GL.End();

				GL.UseProgram(0);
				textureRight.Unbind(); // unbind in reverse order, to have texture unit 0 active afterwards
				textureLeft.Unbind();

				if (ShowCaption)
				{
					// Show caption in white with black border.
					using (OpenTK.Graphics.TextPrinter textPrinter = new OpenTK.Graphics.TextPrinter())
					{
						// Calculate size of caption one and add some slack.
						var extents = textPrinter.Measure(caption, Font);
						var textWidth  = extents.BoundingBox.Width + 1;
						var textHeight = extents.BoundingBox.Height + 1;

						// First print black text for border, then white text centred on top.
						textPrinter.Begin();
						textPrinter.Print(caption, Font, Color.Black, new RectangleF(1, 1, textWidth, textHeight));
						textPrinter.Print(caption, Font, Color.Black, new RectangleF(2, 1, textWidth, textHeight));
						textPrinter.Print(caption, Font, Color.Black, new RectangleF(3, 1, textWidth, textHeight));
						textPrinter.Print(caption, Font, Color.Black, new RectangleF(1, 2, textWidth, textHeight));
						textPrinter.Print(caption, Font, Color.Black, new RectangleF(2, 2, textWidth, textHeight));
						textPrinter.Print(caption, Font, Color.Black, new RectangleF(3, 2, textWidth, textHeight));
						textPrinter.Print(caption, Font, Color.Black, new RectangleF(1, 3, textWidth, textHeight));
						textPrinter.Print(caption, Font, Color.Black, new RectangleF(2, 3, textWidth, textHeight));
						textPrinter.Print(caption, Font, Color.Black, new RectangleF(3, 3, textWidth, textHeight));
						textPrinter.Print(caption, Font, Color.White, new RectangleF(2, 2, textWidth, textHeight));
						textPrinter.End();
					}
				}

				// restore depth buffer state
				GL.DepthMask(true);
				GL.Enable(EnableCap.DepthTest);

				textureLeft.Free();
				textureRight.Free();
			}

			// show results
			SwapBuffers();
		}

		/// <summary>Fits the texture into the target width/height.</summary>
		/// <param name="tex">The input video texture.</param>
		/// <param name="targetWidth">Target Width.</param>
		/// <param name="targetHeight">Target height.</param>
		private void ComputeTargetSize(Texture tex, out float targetWidth, out float targetHeight)
		{
			targetWidth  = Width;
			targetHeight = Height;

			if (PreserveAspectRatio)
			{
				// fit to width if input image is wider than output image
				if (tex.Width * targetHeight >= targetWidth * tex.Height) // multiplied out numerator
					targetHeight = (float)tex.Height / tex.Width * targetWidth;
				else // fit to height if input image is higher than output image
					targetWidth = (float)tex.Width / tex.Height * targetHeight;
			}

			// normalise to get vertex coordinates
			targetWidth  /= Width;
			targetHeight /= Height;
		}

		#endregion // Rendering functions
//=================================================================================================
		#region Event handling + designer view

		private void StereoVideoControl_Load(object sender, EventArgs e)
		{
			if (!InDesignMode)
				InitGL();
		}

		private void StereoVideoControl_Paint(object sender, PaintEventArgs e)
		{
			// custom film strip look for Designer
			if (InDesignMode)
			{
				e.Graphics.Clear(BackColor);
				int framePadding = 4;
				int holePadding = 6;
				int holeSize = 10;

				// black background
				e.Graphics.FillRectangle(new SolidBrush(Color.Black), ClientRectangle);

				// punched holes
				for (int y = 4; y < Height; y += 2 * holeSize)
				{
					e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), holePadding, y, holeSize, holeSize);
					e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), Width - holePadding - holeSize, y, holeSize, holeSize);
				}

				// frames
				int frameBoundary1 = (int)(0.1 * Height + 0.5);
				int frameBoundary2 = (int)(0.9 * Height + 0.5);
				e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), 2 * holePadding + holeSize, 0, Width - 4 * holePadding - 2 * holeSize, frameBoundary1 - framePadding / 2);
				e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), 2 * holePadding + holeSize, frameBoundary1 + framePadding / 2, Width - 4 * holePadding - 2 * holeSize, frameBoundary2 - frameBoundary1 - framePadding);
				e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), 2 * holePadding + holeSize, frameBoundary2 + framePadding / 2, Width - 4 * holePadding - 2 * holeSize, Height - (frameBoundary2 + framePadding / 2));

				// "Video Panel" text
				Font font = new Font(Font.FontFamily, 16, FontStyle.Bold);
				String text = "New OpenGL" + Environment.NewLine + "Stereo Video" + Environment.NewLine + "Panel";
				StringFormat stringFormat = new StringFormat { Alignment = StringAlignment.Center, LineAlignment = StringAlignment.Center };
				e.Graphics.DrawString(text, font, new SolidBrush(ForeColor), Width / 2, Height / 2, stringFormat);
			}
			else
			{
				MakeCurrent();
				Render();
			}
		}

#endregion // Event handling + designer view
//=================================================================================================
	}
}
