﻿// $Id: ExponentialAverageFilter.cs 494 2010-11-15 20:31:09Z cr333 $
using System.Collections.Generic;
using System.ComponentModel;
using OpenTK.Graphics.OpenGL;

namespace VideoLib.OpenGL
{
	/// <summary>Exponentially averages the incoming frames using avg' = w * in + (1 - w) * avg.</summary>
	public class ExponentialAverageFilter : FboNode
	{
		/// <summary>Gets or sets the weight given to the current frame.</summary>
		[Uniform(), Description("Averaging weight given to the current frame")]
		public float Weight { get; set; }


		/// <summary>Initialises a new instance of the <see cref="ExponentialAverageFilter"/> class.</summary>
		public ExponentialAverageFilter()
		{
			Weight = 0.2f;

			AddDefaultPins();
		}

		/// <summary>Processes some frames, to produce new frames.</summary>
		public override void Process()
		{
			// get latest frames
			var frames = GetInputFrames();
			if (PassOnEndOfStreamFrame(frames)) return;
			var textures = CastFramesTo<TextureFrame>(frames);

			// draw quads
			var result = Draw(textures);
			PushFrame(result);

			// free previous frame
			if (previousAverage != null)
				previousAverage.RefDec();

			// keep the result for the next frame
			previousAverage = result;

			// clean up
			textures.RefDec();
		}

		private TextureFrame previousAverage;

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.ExponentialAverage.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Pin = "in", Uniform = "current",  Unit = 0 },
				new TextureBinding { Pin = null, Uniform = "previous", Unit = 1 },
			};

			base.Init();
		}

		/// <summary>Draws to the framebuffer object using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <returns>The resulting texture.</returns>
		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			// allocate a texture that is identical to the input
			var resultTexture = TextureFrame.CreateToMatch(inputs["in"]);
			resultTexture.Alloc();

			// bind textures
			if (previousAverage != null) previousAverage.Bind(TextureUnit.Texture1);

			// for first frame, use weight = 1 to ignore previous average (as it is undefined)
			float w = 0;
			if (previousAverage == null) { w = Weight; Weight = 1; }

			// bind textures, update uniforms and draw to FBO
			Draw(inputs, resultTexture);

			// return to actual weight for next frame
			if (previousAverage == null) Weight = w;

			return resultTexture;
		}
	}
}
