﻿// $Id: OpenGLWorkQueue.cs 440 2010-10-18 11:49:09Z cr333 $
using System.Collections.Generic;
using System.Threading;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib.Scheduling;

namespace VideoLib.OpenGL
{
	/// <summary>Marker interface to indicate that this node contains OpenGL code.</summary>
	public interface IOpenGLNode
	{
	}

	/// <summary>Implements a work queue just for OpenGL-based nodes, which are marked using the <see cref="IOpenGLNode"/> interface.</summary>
	/// <remarks>Should be applied before any CPU-based work queues.</remarks>
	public class OpenGLWorkQueue : WorkQueue
	{
		/// <summary>Hidden native window, to have access to an OpenGL context.</summary>
		private INativeWindow window;

		/// <summary>An OpenGL context is needed to do anything in OpenGL.</summary>
		private OpenTK.Graphics.IGraphicsContext context;

		/// <summary>Starts all worker threads associated with this work queue.</summary>
		/// <returns>Collection of all started worker threads.</returns>
		public override IEnumerable<Thread> StartWorkers()
		{
			TextureFrame.DeleteTextureHandler = DeleteTexture;

			// start just a single thread, as the CUDA context is thread-specific
			StartWorker(0, "OpenGLWorker");
			return workers;
		}

		/// <summary>Main function of a worker thread. This is where the scheduling happens.</summary>
		/// <param name="data">The thread's worker ID.</param>
		protected override void RunWorker(object data)
		{
			// create a GraphicsContext and make it current for his thread
			window = new NativeWindow();
			context = new OpenTK.Graphics.GraphicsContext(OpenTK.Graphics.GraphicsMode.Default, window.WindowInfo);
			context.MakeCurrent(window.WindowInfo);
			context.LoadAll();

			// use on base implementation
			base.RunWorker(data);
		}

		/// <summary>Handles initialisation of nodes.</summary>
		/// <param name="node">The node to be initialised.</param>
		/// <returns>True, if and only if the node has been initialised.</returns>
		public override bool HandleNodeInitEvent(INode node)
		{
			if (node is IOpenGLNode)  // only initialise OpenGL nodes
				return HandleNodeInitEvent(node, Ready);

			return false; // event not handled
		}

		/// <summary>Handles nodes that become ready to run.</summary>
		/// <param name="node">The ready node.</param>
		/// <returns>True, if and only if the event was handled.</returns>
		public override bool HandleNodeReadyEvent(INode node)
		{
			if (node is IOpenGLNode) // add OpenGL nodes to OpenGL-only queue
				return HandleNodeReadyEvent(node, Scheduler.Profiler, Ready);

			return false; // event not handled
		}

		/// <summary>Schedules a texture for deletion on the main OpenGL thread.</summary>
		/// <param name="textureID">The texture ID.</param>
		protected void DeleteTexture(int textureID)
		{
			lock (Ready)
			{
				Ready.Enqueue(new ActionWorkItem(delegate() { GL.DeleteTexture(textureID); }));
				Monitor.Pulse(Ready);
			}
		}
	}
}