﻿// $Id: OpticalFlowFilter.cs 461 2010-10-25 12:12:07Z cr333 $

using System.ComponentModel;
namespace VideoLib.OpenGL.OpticalFlow
{
	/// <summary>Implements optical flow computation based on Martin Eisemann's Brox et al. code.</summary>
	public class OpticalFlowFilter : Node, IOpenGLNode
	{
		#region Public properties
//=================================================================================================

		/// <summary>Gets or sets the number of outer iterations.</summary>
		[Category("Optical Flow"), DefaultValue(30), Description("The number of outer iterations.")]
		public int Iterations
		{
			get { return iterations; }
			set { iterations = value; needToUpdateParameters = true; }
		}

		/// <summary>Gets or sets the number of Jacobi iterations.</summary>
		[Category("Optical Flow"), DefaultValue(1), DisplayName("Jacobi iterations"), Description("The number of Jacobi iterations.")]
		public int JacobiIterations
		{
			get { return jacobiIterations; }
			set { jacobiIterations = value; needToUpdateParameters = true; }
		}

		/// <summary>Gets or sets the alpha value for the Jacobi iteration.</summary>
		/// <remark>It seems that a large value allows for more rigid transformations.</remark>
		[Category("Optical Flow"), DefaultValue(7.6f), Description("Alpha value for the Jacobi iteration, which influences the rigidity of the flow field.")]
		public float Alpha
		{
			get { return alpha; }
			set { alpha = value; needToUpdateParameters = true; }
		}

		/// <summary>Gets or sets the beta value of the Jacob iteration.</summary>
		/// <remarks>Beta is responsible for the smoothness of the flow field.</remarks>
		[Category("Optical Flow"), DefaultValue(0.05f), Description("Beta value for the Jacobi iteration, which influences the smoothness of the flow field.")]
		public float Beta
		{
			get { return beta; }
			set { beta = value; needToUpdateParameters = true; }
		}

		/// <summary>Gets or sets the fine level.</summary>
		/// <remarks>Cannot be changed after initialisation.</remarks>
		[Category("Optical Flow"), DefaultValue(0), DisplayName("Fine level"), Description("Fine level of the image pyramid."), ReadOnly(true)]
		public int FineLevel
		{
			get { return fineLevel; }
			set { fineLevel = value; }
		}

		/// <summary>Gets or sets the coarse level.</summary>
		/// <remarks>Cannot be changed after initialisation.</remarks>
		[Category("Optical Flow"), DefaultValue(5), DisplayName("Coarse level"), Description("Coarse level of the image pyramid."), ReadOnly(true)]
		public int CoarseLevel
		{
			get { return coarseLevel; }
			set { coarseLevel = value; }
		}

		/// <summary>Gets or sets the level offset.</summary>
		/// <remarks>Cannot be changed after initialisation.</remarks>
		[Category("Optical Flow"), DefaultValue(1), DisplayName("Level offset"), Description("Level offset into the image pyramid."), ReadOnly(true)]
		public int LevelOffset
		{
			get { return levelOffset; }
			set { levelOffset  = value; }
		}

//=================================================================================================
		#endregion // Public properties

		private OpticalFlow opticalFlow;

		private bool needToUpdateParameters;
		private int iterations;
		private int jacobiIterations;
		private int fineLevel;
		private int coarseLevel;
		private int levelOffset;
		private float alpha;
		private float beta;

		/// <summary>Initialises a new instance of the <see cref="OpticalFlowFilter"/> class.</summary>
		public OpticalFlowFilter() : base()
		{
			// default parameters
			Iterations       = 30;
			JacobiIterations = 1;
			FineLevel        = 0;
			CoarseLevel      = 5;
			LevelOffset      = 1;
			Alpha            = 7.6f;
			Beta             = 0.05f;

			AddInputPins("in0", "in1");
			AddOutputPin("out");
		}

		/// <summary>Processes some frames, to produce new frames.</summary>
		public override void Process()
		{
			// get the input textures
			var frames = GetInputFrames();
			if (PassOnEndOfStreamFrame(frames)) return;
			var textures = CastFramesTo<TextureFrame>(frames);
			AssertFrameProperties(textures, FrameProps.Size);

			// initialise optical flow
			if (opticalFlow == null)
			{
				opticalFlow = new OpticalFlow() { FineLevel = FineLevel, CoarseLevel = CoarseLevel, LevelOffset = LevelOffset };
				opticalFlow.Init(textures["in0"].Width, textures["in0"].Height);
				needToUpdateParameters = true;
			}

			// update parameters if necessary
			if (needToUpdateParameters)
			{
				opticalFlow.Iterations       = Iterations;
				opticalFlow.JacobiIterations = JacobiIterations;
				opticalFlow.Alpha            = Alpha;
				opticalFlow.Beta             = Beta;
			}

			// prepare a new texture frame to render into
			TextureFrame flow = new TextureFrame(textures["in0"].Width, textures["in0"].Height)
			{
				SequenceNumber = textures["in0"].SequenceNumber
			};
			flow.Texture.Alloc();

			// compute optical flow
			opticalFlow.Compute(textures["in0"].Texture, textures["in1"].Texture, flow.Texture);
			PushFrame(flow);

			// clean up
			textures["in0"].RefDec();
			textures["in1"].RefDec();
			flow.RefDec();
		}
	}
}