﻿// $Id: UndistortFilter.frag 547 2012-01-11 16:04:00Z cr333 $
#version 120
#extension GL_EXT_gpu_shader4 : enable

// input texture
uniform sampler2D input;

// parameters
uniform vec2  fc; // focal lengths
uniform vec2  cc; // principal point
uniform float ac; // skew
uniform float[5] kc; // distortion coefficients


// Computes the determinant of the 3x3 matrix m.
float determinant_3x3(mat3 m)
{
	return m[0].x * (m[1].y * m[2].z - m[2].y * m[1].z)
	     + m[1].x * (m[2].y * m[0].z - m[0].y * m[2].z)
	     + m[2].x * (m[0].y * m[1].z - m[1].y * m[0].z);
}

// Computes the adjoint of the 3x3 matrix m.
mat3 adjoint_3x3(mat3 m)
{
	return mat3(
		m[1].y * m[2].z - m[2].y * m[1].z,
		m[2].y * m[0].z - m[0].y * m[2].z,
		m[0].y * m[1].z - m[1].y * m[0].z,
		m[2].x * m[1].z - m[1].x * m[2].z,
		m[0].x * m[2].z - m[2].x * m[0].z,
		m[1].x * m[0].z - m[0].x * m[1].z,
		m[1].x * m[2].y - m[2].x * m[1].y,
		m[2].x * m[0].y - m[0].x * m[2].y,
		m[0].x * m[1].y - m[1].x * m[0].y
	);
}

// Computes the inverse of the 3x3 matrix m.
mat3 invert_3x3(mat3 m)
{
	return adjoint_3x3(m) / determinant_3x3(m);
}

// Computes image coordinates to compensate lens distortion.
//
// inputs:
//   - intrinsic camera parameters (via uniforms)
//   - coordinates of ideal pin-hole camera
//
// returns:
//   - 2D image coordinates in the distorted input image
vec2 CalculateDistortion(vec2 coords)
{
	// The code in this function has been directly translated from
	// Jean-Yves Bouguet's Camera Calibration Toolbox for Matlab,
	// and then inlined and optimised a little bit.
	//
	// The original files are:
	//   - camera_calib_toolbox/undistort_image.m
	//   - camera_calib_toolbox/rect.m
	//   - camera_calib_toolbox/apply_distortion.m

	// Compute the new KK matrix to fit as much data in the image
	// (in order to accomodate large distortions)
	mat3 KK = mat3(fc.x, 0.0, 0.0, ac * fc.x, fc.y, 0.0, cc.x, cc.y, 1.0); // column-major constructor

	// Matlab: [I2] = rect(I,(R=)eye(3),fc,cc,kc,KK_new);
	// diving into camera_calib_toolbox/rect.m ...

	// Matlab: rays = inv(KK_new)*[(px - 1)';(py - 1)';ones(1,length(px))]; -> 3 x n
	//vec3 rays = inverse(KK) * vec3(coords, 1.0); // supported from #version 150
	vec3 rays = invert_3x3(KK) * vec3(coords, 1.0);

	// Matlab: x = [rays2(1,:)./rays2(3,:);rays2(2,:)./rays2(3,:)]; -> 2 x n
	vec2 xx = rays.xy / rays.z;

	// Matlab: xd = apply_distortion(x,k);
	// diving into camera_calib_toolbox/apply_distortion.m ...

	// Radial distortion
	float r2 = xx.x * xx.x + xx.y * xx.y;
	vec2 xd1 = xx * (1.0 + kc[0] * r2 + kc[1] * r2 * r2 + kc[4] * r2 * r2 * r2);

	// Tangential distortion
	float a1 =      2.0 * xx.x * xx.y;
	float a2 = r2 + 2.0 * xx.x * xx.x;
	float a3 = r2 + 2.0 * xx.y * xx.y;

	vec2 delta_x = vec2(kc[2] * a1 + kc[3] * a2, kc[2] * a3 + kc[3] * a1);
	vec2 xd = xd1 + delta_x;
	
	// returning to camera_calib_toolbox/rect.m ...

	// reconvert to pixels
	return fc * vec2(xd.x + ac * xd.y, xd.y) + cc;
}


void main()
{
	// look up texture size once and reuse it later
	vec2 size = vec2(textureSize2D(input, 0));

	// Note: The half-pixel shift is used to produce results that are as close
	//       as possible to the Matlab reference implementation.

	// calculate the image coordinates which undistort the input image
	vec2 undistortedXY = CalculateDistortion(gl_FragCoord.xy - vec2(0.5));

	// Read the pixel value using bilinear interpolation.
	gl_FragColor = texture2D(input, (undistortedXY + vec2(0.5f)) / size);
}