﻿// $Id: ResamplingFilter.cs 436 2010-10-15 16:26:38Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.Frames;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace VideoLib.Filters
{
	/// <summary>Defines how frames are resampled.</summary>
	public enum ResamplingMode
	{
		/// <summary>Nearest-neighbour interpolation is the fastest, but roughest technique.</summary>
		NearestNeighbour,

		/// <summary>Bilinear interpolation provides a good tradeoff between speed and accuracy.</summary>
		Bilinear,

		/// <summary>Bicubic interpolation provides the best interpolation results available here.</summary>
		Bicubic
	}

	/// <summary>A filter that can be used for up- or downsampling of bitmap frames.</summary>
	public class ResamplingFilter : Node
	{
		/// <summary>Target width.</summary>
		public int Width { get; set; }

		/// <summary>Target height.</summary>
		public int Height { get; set; }

		/// <summary>Interpolation technique to use (default: bilinear).</summary>
		public ResamplingMode Mode { get; set; }

		/// <summary>If true, will try to preserve the aspect ratio of the input frames. Default: true.</summary>
		public bool PreserveAspectRatio { get; set; }

		/// <summary>Constructs a new resampling filter using bilinear interpolation.</summary>
		/// <param name="targetWidth">Target width.</param>
		/// <param name="targetHeight">Target height.</param>
		public ResamplingFilter(int targetWidth, int targetHeight) : base()
		{
			Mode = ResamplingMode.Bilinear;
			Width = targetWidth;
			Height = targetHeight;
			PreserveAspectRatio = true;

			AddInputPin("in");
			AddOutputPin("out");
		}

		/// <summary>Constructs a new resampling filter using bilinear interpolation.</summary>
		/// <param name="targetSize">Target size.</param>
		public ResamplingFilter(Size targetSize)
			: this(targetSize.Width, targetSize.Height)
		{
		}

		/// <summary>Processes a frame.</summary>
		public override void Process()
		{
			// get input bitmap
			Frame frame = GetInputFrame("in");
			if (PassOnEndOfStreamFrame(frame)) return;
			BitmapFrame inputFrame = CastFrameTo<BitmapFrame>(frame);

			// bypass resampling if the resolution doesn't change
			if (Width == inputFrame.Width && Height == inputFrame.Height)
			{
				PushFrame("out", inputFrame.Clone());
				return;
			}

			// apply the resampling and pass on result
			BitmapFrame outputFrame = BuiltInResampling(inputFrame);
			PushFrame("out", outputFrame);
		}

		/// <summary>Resamples a frame using Graphics.DrawImage.</summary>
		/// <param name="inputFrame">Input frame.</param>
		/// <returns>Resampled output frame.</returns>
		private BitmapFrame BuiltInResampling(BitmapFrame inputFrame)
		{
			Bitmap inputBitmap = inputFrame.Bitmap;

			// compute target width and height if asked to preserve aspect ratios
			int targetWidth = Width;
			int targetHeight = Height;
			if (PreserveAspectRatio)
			{
				// scale to fit width
				targetWidth = Width;
				targetHeight = (int)(0.5 + (double)inputBitmap.Height / (double)inputBitmap.Width * targetWidth);

				// scale to fit height
				if (targetHeight > Height)
				{
					targetHeight = Height;
					targetWidth  = (int)(0.5 + (double)inputBitmap.Width / (double)inputBitmap.Height * targetHeight);
				}
			}

			// if no resizing required, just return input frame
			if (inputBitmap.Width == targetWidth && inputBitmap.Height == targetHeight)
				return inputFrame.Clone() as BitmapFrame;

			// create output bitmap
			Bitmap outputBitmap = new Bitmap(targetWidth, targetHeight, inputBitmap.PixelFormat);
			Graphics outputGraphics = Graphics.FromImage(outputBitmap);

			// set interpolation mode
			switch (Mode)
			{
				case ResamplingMode.NearestNeighbour: outputGraphics.InterpolationMode = InterpolationMode.NearestNeighbor; break;
				case ResamplingMode.Bilinear: outputGraphics.InterpolationMode = InterpolationMode.HighQualityBilinear; break;
				case ResamplingMode.Bicubic: outputGraphics.InterpolationMode = InterpolationMode.HighQualityBicubic; break;
				default: throw new Exception(String.Format("Unsupported interpolation mode \"{0}\".", Mode));
			}

			// use Graphics.DrawImage to do the resampling for us
			
			// downsampling
			if (targetWidth < inputBitmap.Width || targetHeight < inputBitmap.Height)
			{
				//outputGraphics.DrawImage(inputBitmap, -0.5f, -0.5f, targetWidth + 1.0f, targetHeight + 1.0f);
				outputGraphics.DrawImage(inputBitmap,
					new RectangleF(-0.5f, -0.5f, targetWidth + 1, targetHeight + 1),
					new RectangleF(0, 0, inputBitmap.Width, inputBitmap.Height), GraphicsUnit.Pixel);
			}
			else // upsampling
			{
				//float pixelSizeX = (float)targetWidth / inputBitmap.Width;
				//float pixelSizeY = (float)targetHeight / inputBitmap.Height;
				outputGraphics.DrawImage(inputBitmap,
					new RectangleF(0, 0, targetWidth, targetHeight),
					new RectangleF(-0.5f, -0.5f, inputBitmap.Width, inputBitmap.Height), GraphicsUnit.Pixel);
			}

			outputGraphics.Dispose();

			// construct a frame from the resampled bitmap, preserving the sequence number
			return new BitmapFrame(outputBitmap) { SequenceNumber = inputFrame.SequenceNumber };
		}
	}
}
