﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace VideoLib.Parameters
{
    /// <summary>
    /// An experimental Control which provides graphic editing functionalities to an object's 
    /// properties that are marked by the custom attribute "ParameterAttribute"
    /// 
    /// Currently only boolean and int type parameters are recognized. 
    /// </summary>
    public partial class ParameterControl : UserControl
    {
		/// <summary>
		/// Initializes a new instance of the <see cref="ParameterControl"/> class.
		/// </summary>
        public ParameterControl()
        {
            InitializeComponent();
        }

        private List<AbstractParameterControl> ParamControls = new List<AbstractParameterControl>();

        /// <summary>
        /// Initialise this control and link it with all annotated properties within the given variable.  
        /// </summary>
        /// <param name="var"></param>
        public void LinkWithVariable(Object var)
        {
            Type VarType = var.GetType();
            List<ParameterInfo> ParamList = ParameterReflector.EnumerateParameters(VarType);


            if (ParamControls.Count > 0)
            {
                foreach (AbstractParameterControl pc in ParamControls)
                    pc.DisposeControl();
                ParamControls.Clear();
            }

            int Y = 0;
            

            foreach(ParameterInfo pi in ParamList)
            {
                AbstractParameterControl ctrl = ParameterControlFactory.CreateParamControl(pi, var);
                Y += ctrl.CreateControl(this, 0, Y).Height;
            }

            Size = new Size(Width, Y);
        }

		/// <summary>
		/// Handles paint events, for a custom look in the Designer.
		/// </summary>
		/// <param name="e">A <see cref="T:System.Windows.Forms.PaintEventArgs"/> that contains the event data.</param>
        protected override void OnPaint(PaintEventArgs e)
        {
            // custom look for Designer
            if (DesignMode)
            {
                Font font = new Font("Arial", 24, FontStyle.Bold, GraphicsUnit.Pixel);
                e.Graphics.DrawString("Parameter Control", font, Brushes.Black, 0, 0);
            }
        }
    }

    /// <summary>
    /// Base class for a sub control. Each instance is mapped to one recognized parameter entry.
    /// </summary>
    abstract class AbstractParameterControl
    {
        protected Control[] controls = null;
        protected string ParamName;
        protected ValueSetter ParamSetter;

        // Create UI, returns the aggregated size of this control
        abstract public Size CreateControl(Control Parent, int X, int Y);

        public void DisposeControl()
        {
            foreach(Control c in controls) 
                if (c != null)
                    c.Dispose();
            controls = null;
        }
    }

    
    delegate Boolean ValueSetter(Object x);
    
    // Boolean parameter - Simply a CheckBox
    class BooleanParameterControl : AbstractParameterControl
    {
        protected Boolean ParamValue;

        public BooleanParameterControl(String Name, Boolean DefaultVal, ValueSetter setter)
        {
            ParamName = Name;
            ParamValue = DefaultVal;
            ParamSetter = setter;
            setter(DefaultVal);
        }

        public override Size CreateControl(Control Parent, int X, int Y)
        {
            if (controls != null)
                DisposeControl();

            CheckBox chk = new CheckBox();
            chk.Text = ParamName;
            chk.Checked = ParamValue;
            chk.Location = new Point(X, Y);
            chk.AutoSize = true;
            chk.Visible = true;
            Parent.Controls.Add(chk);
            chk.CheckedChanged += new EventHandler(this.CheckBox_CheckedChanged);

            controls = new Control[1];
            controls[0] = chk;

            return chk.Size;
        }

        private void CheckBox_CheckedChanged(Object sender, EventArgs e)
        {
            ParamValue = ((CheckBox)sender).Checked;
            ParamSetter(ParamValue);
        }
    }

    // Int parameter - A labeled TextBox if its value range is not specified 
    // A labeled ScrollBar if the range is specified
    class IntParameterControl : AbstractParameterControl
    {
        protected int ParamValue;
        protected Boolean RangeLimited;
        protected int MinVal, MaxVal;

        ///Constructor for free range int parameter
        public IntParameterControl(String Name, int DefaultVal, ValueSetter setter)
        {
            ParamName = Name;
            ParamValue = DefaultVal;
            ParamSetter = setter;
            setter(DefaultVal);
            RangeLimited = false;
        }
        //Constructor for limited -range int parameter.
        public IntParameterControl(String Name, int DefaultVal, int MinVal, int MaxVal, ValueSetter setter)
        {
            ParamName = Name;
            ParamValue = DefaultVal;
            ParamSetter = setter;
            setter(DefaultVal);
            RangeLimited = true;
            this.MinVal = MinVal;
            this.MaxVal = MaxVal;
        }

        public override Size CreateControl(Control Parent, int X, int Y)
        {
            if (controls != null)
                DisposeControl();

            controls = new Control[2];

            Label lbl = new Label();
            lbl.Text = ParamName;
            lbl.AutoSize = true;
            lbl.Location = new Point(X, Y);
            Parent.Controls.Add(lbl);
            controls[0] = lbl;

            int ControlHeight = 0;
            if (RangeLimited)
            {
                HScrollBar scroll = new HScrollBar();
                scroll.Minimum = MinVal;
                scroll.Maximum = MaxVal;
                scroll.Value = ParamValue;
                scroll.Size = new Size(Parent.Width - lbl.Width - X, lbl.Height);
                scroll.Location = new Point(X + lbl.Width, Y);
                ControlHeight = scroll.Height;
                Parent.Controls.Add(scroll);
                scroll.ValueChanged += new System.EventHandler(this.Scroll_Changed);
                controls[1] = scroll;
            }else
            {
                TextBox txt = new TextBox();
                txt.Text = ParamValue.ToString();
                txt.Size = new Size(Parent.Width - lbl.Width - X, lbl.Height);
                txt.Location = new Point(X, Y);
                txt.Location = new Point(X + lbl.Width, Y);
                ControlHeight = txt.Height;
                Parent.Controls.Add(txt);
                txt.TextChanged += new System.EventHandler(this.Text_Changed);
                controls[1] = txt;
            }

            return new Size(Parent.Width, Math.Max(ControlHeight, lbl.Height));
        }

        private void Scroll_Changed(Object sender, EventArgs e)
        {
            ParamValue = ((HScrollBar)sender).Value;
            ParamSetter(ParamValue);
        }
        private void Text_Changed(Object sender, EventArgs e)
        {
            int t;
            if (Int32.TryParse(((TextBox)sender).Text, out t))
                ParamValue = t;
            ParamSetter(ParamValue);
        }
    }

    // Class Factory for AbstractParameterControl
    class ParameterControlFactory
    {
        public static AbstractParameterControl CreateParamControl(ParameterInfo pi, Object obj)
        {
            AbstractParameterControl ctrl;
            if (pi.ParamType.Equals(typeof(Boolean)))
                ctrl = new BooleanParameterControl(pi.Name, (Boolean)pi.GetValue(obj),
                    x => { pi.SetValue(obj, x); return true; }
                    );
            else if (pi.ParamType.Equals(typeof(int)))
            {
                if (  (pi.Maximum != null) && (pi.Minimum != null))
                    ctrl = new IntParameterControl(pi.Name, (int)pi.GetValue(obj),
                        (int)pi.Minimum, (int)pi.Maximum,
                        x => { pi.SetValue(obj, x); return true; }
                        );
                else
                    ctrl = new IntParameterControl(pi.Name, (int)pi.GetValue(obj),
                        x => { pi.SetValue(obj, x); return true; }
                        );
            }
                else
                    ctrl = null;

            return ctrl;
        }
    }
    


}
