﻿using System;
using System.Collections.Generic;

namespace VideoLib.Parameters
{
    /// <summary>
    /// A simple collection of settings, stored on a per-declared-type basis (i.e. separately for each type in an inheritance hierarchy),
    /// and can be applied when required
    /// </summary>
    [Serializable]
    public class ParameterValueDictionary
    {
        /// <summary>
        /// A single setting for a specific parameter
        /// </summary>
        [Serializable]
        public class ParameterSetting
        {
            /// <summary>
            /// The parameter represented
            /// </summary>
            public ParameterInfo param;

            /// <summary>
            /// The saved value
            /// </summary>
            public object value;

            /// <summary>
            /// Creates and initializes a <see cref="ParameterSetting"/>
            /// </summary>
            /// <param name="paramInfo">Describes the parameter to be referred to</param>
            public ParameterSetting(ParameterInfo paramInfo)
            {
                param = paramInfo;
                value = paramInfo.Default;
            }

            /// <summary>
            /// Checks for equality of the underlying ParameterInfo object
            /// </summary>
            /// <param name="obj">Another object to compare to</param>
            /// <returns><c>true</c> if obj and this instance are the same type and wrap the same parameter, otherwise <c>false</c></returns>
            public override bool Equals(object obj)
            {
                if (!(obj is ParameterSetting))
                    return false;
                return ((ParameterSetting)obj).param == param;
            }

            /// <summary>
            /// Return a hash code consistent with <see cref="Equals"/>
            /// </summary>
            /// <returns>A 32-bit signed integer that is the hash code for this instance</returns>
            public override int GetHashCode()
            {
                return param.GetHashCode();
            }
        }

        /// <summary>
        /// The dictionary used to store all of the settings in-memory, indexed by type
        /// </summary>
        private readonly Dictionary<Type, List<ParameterSetting>> typeSettings;

        /// <summary>
        /// Create an empty instance of <see cref="ParameterValueDictionary"/>
        /// </summary>
        public ParameterValueDictionary()
        {
            typeSettings = new Dictionary<Type,List<ParameterSetting>>();
        }

        /// <summary>
        /// Retrieves the settings for the specified type from the dictionary, or creates them, and returns the settings
        /// </summary>
        /// <param name="t">The type of the target of the settings</param>
        /// <returns>A populated list of parameter settings for type <paramref name="t"/></returns>
        private List<ParameterSetting> RetrieveOrAddType(Type t)
        {
            List<ParameterSetting> tSettings;
            if (!typeSettings.TryGetValue(t, out tSettings))
            {
                // The type has not been used before: add it to the dictionary
                tSettings = new List<ParameterSetting>();

                // Add all of the parameters for that type to the list
                List<ParameterInfo> parameters = ParameterReflector.EnumerateParameters(t, false);
                foreach (ParameterInfo pInfo in parameters)
                    tSettings.Add(new ParameterSetting(pInfo));

                typeSettings.Add(t, tSettings);
            }
            return tSettings;
        }

        /// <summary>
        /// Saves the current settings for the specified object to the in-memory dictionary
        /// </summary>
        /// <param name="obj">The object for which to store the current 'parameters state'</param>
        public void StoreSettings(object obj)
        {
            Type concreteType = obj.GetType();

            // Iterate up the inheritance hierarchy, setting the parameters for each type individually
            for (Type t = concreteType; t != null; t = t.BaseType)
            {
                List<ParameterSetting> tSettings = RetrieveOrAddType(t);

                // Set the stored settings to the current state of the object passed
                for(int i = 0; i < tSettings.Count; ++i)
                    tSettings[i].value = tSettings[i].param.GetValue(obj);
            }
        }
        /// <summary>
        /// Loads the stored settings for the type hierarchy of <paramref name="obj"/>, and applies them to
        /// <paramref name="obj"/>
        /// </summary>
        /// <param name="obj">The object to set up from the stored state settings</param>
        public void ApplySettings(object obj)
        {
            Type concreteType = obj.GetType();

            // Copy the type hierarchy information into a linked list, in reverse order
            LinkedList<Type> hierarchy = new LinkedList<Type>();
            for (Type t = concreteType; t != null; t = t.BaseType)
                hierarchy.AddFirst(t);

            // Iterate down the inheritance hierarchy, applying the settings for each type individually
            // (from most general to most specific, in order to properly handle overridden parameters)
            foreach(Type t in hierarchy)
            {
                List<ParameterSetting> tSettings = RetrieveOrAddType(t);

                // Set the object parameter settings to those found
                for (int i = 0; i < tSettings.Count; ++i)
                    tSettings[i].param.SetValueLimited(obj, tSettings[i].value);
            }
        }
    }
}
