﻿// $Id: BaselineScheduler.cs 436 2010-10-15 16:26:38Z cr333 $
// Based on a Part II project by Rubin Xu (2010).
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using VideoLib.Frames;
using System.Diagnostics;
using VideoLib.Utils;

namespace VideoLib.Scheduling
{
	/// <summary>Implements a baseline scheduler with one thread per node.</summary>
	public class BaselineScheduler : Scheduler
	{
		/// <summary>Constructs a baseline scheduler.</summary>
		/// <param name="topology">Topology of nodes to use.</param>
		public BaselineScheduler(Topology topology) : base(topology)
		{
			Status = SchedulerStatus.Offline;
		}

		/// <summary>Sets up the scheduler, preparing it for the work ahead.</summary>
		public override void Setup()
		{
			base.Setup();

			try
			{
				// Initialise state
				Status = SchedulerStatus.Starting;

				// register NewFrameEventListener to listen for all newly produced frames
				foreach (var edge in Topology.OutgoingEdges)
				{
					OutputPin outputPin = edge.Key;
					outputPin.NewFrame += new NewFrameEventHandler(NewFrameEventListener);
				}

				// register NodeStatusChangedListener with all nodes, to listen to all node status changes
				foreach (INode node in Topology.Nodes)
					node.StatusChanged += new NodeStatusChangedHandler(NodeStatusChangedListener);

				// start up all nodes by initialising them and updating their status
				foreach (INode node in Topology.Nodes)
				{
					node.Init();
					node.UpdateStatus();
				}

				// Initialization done, switch to running state
				Status = SchedulerStatus.Running;
			}
			catch (Exception)
			{
				throw;
			}
		}

		/// <summary>Dummy implementation of RunWorker.</summary>
		/// <param name="data">The thread's worker ID, if <see cref="Scheduler.StartWorker"/> is not not overidden.</param>
		protected override void RunWorker(object data)
		{
			throw new NotImplementedException();
		}

		///<summary>
		/// Dummy argument method to start all workers. As the baseline scheduler
		/// spawns a lot of threads and let OS to do the scheduling, there is no need
		/// to specify how many workers are needed         
		/// </summary>
		/// <param name="NumWorkers">TDummy argument</param>
		public override void StartWorkers(int NumWorkers)
		{
			int nextWorkerThread = 1;
			foreach (Node node in Topology.Nodes)
			{
				Thread worker = new Thread(RunWorkerThread) { Name = String.Format("Worker{0}", nextWorkerThread) };
				nextWorkerThread++;
				worker.Start(node);
				workers.Add(worker);
			}
		}
		/// <summary>Main function of a worker thread. One worker per Node</summary>
		protected void RunWorkerThread(Object data)
		{
			// Run an infinite loop, looking for work to do
			INode node = (INode)data;
			while (Status == SchedulerStatus.Running)
			{
				try
				{
					lock (node)
					{                        
						// Wait until the node becomes Ready
						// ( otherwise block until someone pulses it (from NodeStatusChangedListener) *)
						while (node.Status != NodeStatus.Ready &&
							   Status == SchedulerStatus.Running)
						{
							Monitor.Wait(node, 1000);
						}
						if (Status == SchedulerStatus.Running)
						{
							Debug.WriteLine(String.Format("{0}: Node<{1}> processing", Thread.CurrentThread.Name, node.ToString()));
							Profiler.MarkStarting(node);
							node.PreProcess();
							node.Process();
							node.PostProcess();
							Profiler.MarkEnding(node);
						}
					}
					Thread.Sleep(0); //Yield
				}
				catch (ThreadInterruptedException)
				{
					break;
				}
			}

			Console.WriteLine(Thread.CurrentThread.Name + ": Stopped.");
		}

		// Propagate newFrame event to all inputPins
		private void NewFrameEventListener(object sender, NewFrameEventArgs e)
		{
			OutputPin outputPin = (OutputPin)sender;
			foreach (var inputPin in Topology.OutgoingEdges[outputPin])
			{
				inputPin.NewFrameListener(outputPin, e);
				Debug.WriteLine(String.Format("Transport: {0} => {1}", e.Frame.GetType().Name, outputPin));
			}
		}

		// Listen for status change, notify worker thread if required.
		private void NodeStatusChangedListener(object sender, NodeStatusChangedEventArgs e)
		{
			Debug.WriteLine(String.Format("{0}: {1} changed: {2} => {3}", Thread.CurrentThread.Name, sender, e.OldStatus, e.NewStatus));
			INode node = (INode)sender;

			// if a node is now ready to run, signal it.
			if (e.NewStatus == NodeStatus.Ready)
			{
				Monitor.Pulse(node);
				return;
			}

			// if a node is now waiting for data, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Waiting)
			{
				return;
			}

			// if a node is was ready and is now running, there is nothing we need to do
			if (e.OldStatus == NodeStatus.Ready && e.NewStatus == NodeStatus.Running)
			{
				return;
			}

			// if a node is finished with processing, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Finished)
			{
				return;
			}

			// if a node is now stopped, exit everything
			// TODO: do something more graceful here ...
			if (e.NewStatus == NodeStatus.Stopped)
			{
				foreach (Thread worker in workers)
					worker.Interrupt();
				Console.Error.WriteLine("Node " + node.Name + " was stopped.");
				Stop();
				return;
			}

			//throw new NotImplementedException(String.Format("Cannot handle state transition {0} => {1} of node {2}.", e.OldStatus, e.NewStatus, node));
		}
	}
}
