﻿// $Id: CudaAwareScheduler.cs 293 2010-08-04 16:26:33Z cr333 $
// Based on a Part II project by Rubin Xu (2010).
// Derived from SimpleDequeScheduler.cs
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using VideoLib.Frames;
using System.Diagnostics;
using VideoLib.Utils;

namespace VideoLib.Scheduling
{
	/// <summary>Implements a CUDA-aware queue-based scheduler.</summary>
	/// <remarks>
	/// This scheduler maintains three queues of work items: one each for CUDA tasks, for CPU tasks
	/// and for asynchronous status updates. CUDA-aware means all nodes implementing ICudaNode are
	/// served by a specific thread only.Worker threads are constantly hungry for new work and run
	/// the front item of the queue. Tasks that are ready to be processed are enqueued at the back,
	/// to run them in-order.
	/// 
	/// IMPORTANT: This scheduler, as well as <see cref="ModularScheduler"/> and
	/// <see cref="SimpleDequeScheduler"/> are the only working schedulers at the moment,
	/// due to changes in how events are handled.
	/// </remarks>
	public class CudaAwareScheduler : Scheduler
	{
		/// <summary>Constructs a CUDA-aware queue-based scheduler.</summary>
		/// <param name="topology">Topology of nodes to use.</param>
		public CudaAwareScheduler(Topology topology) : base(topology)
		{
		}

		/// <summary>Ready queue for CPU-based tasks.</summary>
		private readonly Queue<WorkItem> ReadyCPU  = new Queue<WorkItem>();

		/// <summary>Ready queue for CUDA-based tasks.</summary>
		private readonly Queue<WorkItem> ReadyCUDA = new Queue<WorkItem>();

		/// <summary>Ready queue for status updates and transporting frames.</summary>
		private readonly Queue<WorkItem> ReadyStat = new Queue<WorkItem>();

		/// <summary>Sets up the scheduler, preparing it for the work ahead.</summary>
		public override void Setup()
		{
			base.Setup();
			try
			{
				// initialise status
				Status = SchedulerStatus.Starting;

				// listen to all new frames and status updates from output pins
				foreach (var edgeFan in Topology.OutgoingEdges)
				{
					edgeFan.Key.NewFrame += NewFrameListener;
					edgeFan.Key.StatusChanged += PinStatusChangedListener;
				}

				// listen to all status updates from input pins
				foreach (var edgeFan in Topology.IncomingEdges)
					edgeFan.Key.StatusChanged += PinStatusChangedListener;

				// listen to all status changes from nodes
				foreach (INode node in Topology.Nodes)
					node.StatusChanged += NodeStatusChangedListener;

				// start up all nodes
				foreach (INode node in Topology.Nodes)
				{
					node.Init();
					node.UpdateStatus();
				}

				// let's go over to proper running mode
				Status = SchedulerStatus.Running;
			}
			catch (Exception)
			{
				throw;
			}
		}

		/// <summary>Starts a number of worker threads.</summary>
		/// <param name="numWorkers">Number of CPU worker threads to start.</param>
		public override void StartWorkers(int numWorkers)
		{
			for (int i = -2; i < numWorkers; i++)
				StartWorker(i);
		}

		/// <summary>Starts a new worker thread.</summary>
		/// <param name="workerID">The worker thread ID.</param>
		/// <returns>Thread handle of the created worker thread.</returns>
		protected override Thread StartWorker(int workerID)
		{
			string name = String.Format("Worker{0}", workerID);
			if (workerID == -1) name = "CudaWorker";
			if (workerID == -2) name = "StatWorker";

			// start a new scheduler thread
			Thread worker = new Thread(RunWorker) { Name = name };
			worker.Start(workerID);
			workers.Add(worker);
			return worker;
		}

		/// <summary>Main function of a worker thread. This is where the scheduling happens.</summary>
		/// <param name="data">The thread's worker ID, if StartWorkers not not overidden.</param>
		protected override void RunWorker(object data)
		{
			// for profiling idle time
			long idletime = 0;

			// the corresponding ready queue, to be assigned later
			Queue<WorkItem> Ready = null;

			// local ID of worker thread: -1 is CUDA; 0, 1, 2, ... is for CPU
			int workerID = (int)data;

			if (workerID == -1) // CUDA thread
			{
				Ready = ReadyCUDA;

				// lock current managed thread to underlying OS thread
				SchedulerHelper.PinCurrentThreadToCore(0);
			}
			else if (workerID == -2) // status updater thread
			{
				Ready = ReadyStat;

				// lock current managed thread to underlying OS thread
				SchedulerHelper.PinCurrentThreadToCore(1);
			}
			else // CPU threads
			{
				Ready = ReadyCPU;

				// lock current managed thread to underlying OS thread (given by worker ID)
				SchedulerHelper.PinCurrentThreadToCore(workerID);
			}

			// run an infinite loop, looking for work to do
			while(Status == SchedulerStatus.Running)
			{
				WorkItem workitem = null;

				lock (Ready)
				{
					// check if there is work to do (first chance)
					while (Ready.Count == 0 && Status == SchedulerStatus.Running)
					{
						if (Profiling) idletime -= Stopwatch.GetTimestamp();
						Monitor.Wait(Ready, 100);
						if (Profiling) idletime += Stopwatch.GetTimestamp();
					}

					// check if there is work to do (second chance)
					if (Ready.Count == 0 || Status != SchedulerStatus.Running)
						continue;

					// get work item to run
					workitem = Ready.Dequeue();
				}

				// run the work item
				workitem.Run();
			}

			SchedulerHelper.UnpinCurrentThread();
			if(Profiling) Console.WriteLine(Thread.CurrentThread.Name + ": Stopped. Idle time: {0:F3} s", (double)idletime / Stopwatch.Frequency);
			else          Console.WriteLine(Thread.CurrentThread.Name + ": Stopped.");
		}

		/// <summary>Schedules transport of new frames to receivers.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="VideoLib.Frames.NewFrameEventArgs"/> instance containing the event data.</param>
		private void NewFrameListener(object sender, NewFrameEventArgs e)
		{
			OutputPin outputPin = (OutputPin)sender;
			foreach (var inputPin in Topology.OutgoingEdges[outputPin])
			{
				// increment reference count of CUDA frames
				ICudaFrame cudaFrame = e.Frame as ICudaFrame;
				if (cudaFrame != null) cudaFrame.RefInc();

				lock (ReadyStat)
				{
					ReadyStat.Enqueue(new TransportWorkItem(outputPin, inputPin, e));
					
					// wake up a worker thread to process this new item.
					Monitor.Pulse(ReadyStat);
				}
			}
		}

		/// <summary>Listens to status changes in IO pins.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="VideoLib.PinStatusChangedEventArgs"/> instance containing the event data.</param>
		private void PinStatusChangedListener(object sender, PinStatusChangedEventArgs e)
		{
			if (sender is InputPin) // then update upstream output pins
			{
				foreach (var outPin in Topology.IncomingEdges[(InputPin)sender])
				{
					lock (ReadyStat)
					{
						ReadyStat.Enqueue(new ActionWorkItem(delegate() { outPin.DownStreamStatusChangedListener(sender, e); }));
						Monitor.Pulse(ReadyStat);
					}
				}
				return;
			}

			if (sender is OutputPin) // then update node
			{
				OutputPin outputPin = (OutputPin)sender;

				lock (ReadyStat)
				{
					ReadyStat.Enqueue(new ActionWorkItem(delegate() { outputPin.Node.UpdateStatus(); }));
					Monitor.Pulse(ReadyStat);
				}
			}
		}

		/// <summary>Listens to status changes to enqueue ready nodes.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="VideoLib.NodeStatusChangedEventArgs"/> instance containing the event data.</param>
		private void NodeStatusChangedListener(object sender, NodeStatusChangedEventArgs e)
		{
			Debug.WriteLine(String.Format("{0}: {1} changed: {2} => {3}", Thread.CurrentThread.Name, sender, e.OldStatus, e.NewStatus));
			INode node = (INode)sender;

			// if a node is ready to run, append it to the ready queue
			if (e.NewStatus == NodeStatus.Ready)
			{
				var workItem = new ProcessingWorkItem(node) { Profiler = Profiler };

				if (node is ICudaNode) // add CUDA nodes to CUDA-only queue
				{
					lock (ReadyCUDA)
					{
						ReadyCUDA.Enqueue(workItem);
						Monitor.Pulse(ReadyCUDA);
					}
				}
				else // add items to CPU queue
				{
					lock (ReadyCPU)
					{
						ReadyCPU.Enqueue(workItem);
						Monitor.Pulse(ReadyCPU);
					}
				}

				return;
			}

			// if a node is now waiting for data, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Waiting) return;

			// if a node is was ready and is now running, there is nothing we need to do
			if (e.OldStatus == NodeStatus.Ready && e.NewStatus == NodeStatus.Running) return;

			// if a node is finished with processing, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Finished) return;

			// if a node is now stopped, exit everything
			if (e.NewStatus == NodeStatus.Stopped)
			{
				Console.Error.WriteLine("Node " + node.Name + " was stopped.");
				Stop();
				return;
			}

			throw new NotImplementedException(String.Format("Cannot handle state transition {0} => {1} of node {2}.", e.OldStatus, e.NewStatus, node));
		}
	}
}
