﻿// $Id: SimpleQueueScheduler.cs 436 2010-10-15 16:26:38Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using VideoLib.Frames;
using System.Diagnostics;

namespace VideoLib.Scheduling
{
	/// <summary>Implements a simple queue-based scheduler.</summary>
	public class SimpleQueueScheduler : Scheduler
	{
		/// <summary>Constructs a simple queue-based scheduler.</summary>
		/// <param name="topology">Topology of nodes to use.</param>
		public SimpleQueueScheduler(Topology topology) : base(topology)
		{
			Ready = new Queue<INode>();
			Status = SchedulerStatus.Offline;
		}

		/// <summary>Read queue: queue of nodes that are ready to run.</summary>
		private Queue<INode> Ready { get; set; }

		/// <summary>Sets up the scheduler, preparing it for the work ahead.</summary>
		public override void Setup()
		{
			base.Setup();

			// let's start the scheduler then
			Status = SchedulerStatus.Starting;

			// connect up all nodes according to the topology
			foreach (var edgeFan in Topology.OutgoingEdges)
			{
				OutputPin outputPin = edgeFan.Key;
				foreach (var inputPin in edgeFan.Value)
				{
					// make the input pin listen for new frames from the output pin
					outputPin.NewFrame += new NewFrameEventHandler(inputPin.NewFrameListener);
				}
			}

			// register NodeStatusChangedListener with all nodes, to listen to all node status changes
			foreach (INode node in Topology.Nodes)
				node.StatusChanged += new NodeStatusChangedHandler(NodeStatusChangedListener);

			// start up all nodes by initialising them and updating their status
			foreach (INode node in Topology.Nodes)
			{
				node.Init();
				node.UpdateStatus();
			}

			// let's go over to proper running mode
			Status = SchedulerStatus.Running;
		}

		/// <summary>Main function of a worker thread. This is where the scheduling happens.</summary>
		/// <param name="data">The thread's worker ID, if <see cref="Scheduler.StartWorker"/> is not not overidden.</param>
		protected override void RunWorker(object data)
		{
			// Run an infinite loop, looking for work to do
			while(Status == SchedulerStatus.Running)
			{
				// take first node from Ready queue and run it
				INode node = null;
				lock (Ready)
				{
					// check if there is work to do (first chance)
					if (Ready.Count == 0)
					{
						//Console.WriteLine(Thread.CurrentThread.Name + ": waiting for work.");
						Monitor.Wait(Ready, 100);
						//Console.WriteLine(Thread.CurrentThread.Name + ": woke up.");
					}

					// check if there is work to do (second chance)
					if (Ready.Count == 0)
						continue;

					node = Ready.Dequeue();
				}

				// run the node
				//Console.WriteLine(Thread.CurrentThread.Name + ": running " + node.Name);

				node.PreProcess();
				lock (node)
				{
					node.Process();
				}
				node.PostProcess();
			}

			Console.WriteLine(Thread.CurrentThread.Name + ": Stopped.");
		}

		/// <summary>Prints the current state of the scheduler.</summary>
		public void PrintState()
		{
			Console.WriteLine("Scheduler (" + this.Status.ToString() + "):");
			lock (Ready)
			{
				Console.WriteLine("  Ready: " + String.Join(", ", Ready.Select(node => node.Name + " (" + node.Status + ")").ToArray()));
			}
			Console.WriteLine();
		}

		private void NodeStatusChangedListener(object sender, NodeStatusChangedEventArgs e)
		{
			Debug.WriteLine(String.Format("{0}: {1} changed: {2} => {3}", Thread.CurrentThread.Name, sender, e.OldStatus, e.NewStatus));
			INode node = (INode)sender;

			// if a node is now ready to run, append it to the Ready queue
			if (e.NewStatus == NodeStatus.Ready)
			{
				lock (Ready)
				{
					Ready.Enqueue(node);
					Monitor.Pulse(Ready);
				}
				return;
			}

			// ???
			if (e.OldStatus == NodeStatus.Ready && e.NewStatus == NodeStatus.Waiting)
			{
				return;
			}

			// if a node is now waiting for data, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Waiting)
			{
				return;
			}

			// if a node is was ready and is now running, there is nothing we need to do
			if (e.OldStatus == NodeStatus.Ready && e.NewStatus == NodeStatus.Running)
			{
				return;
			}

			// if a node is finished with processing, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Finished)
			{
				return;
			}

			// if a node is now stopped, exit everything
			// TODO: do something more graceful here ...
			if (e.NewStatus == NodeStatus.Stopped)
			{
				Console.Error.WriteLine("Node " + node.Name + " was stopped.");
				Stop();
				return;
			}

			throw new NotImplementedException(String.Format("Cannot handle state transition {0} => {1} of node {2}.", e.OldStatus, e.NewStatus, node));
		}
	}
}
