﻿// $Id: SimpleQueueScheduler2.cs 436 2010-10-15 16:26:38Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using VideoLib.Frames;

namespace VideoLib.Scheduling
{
	/// <summary>Scheduler that schedules both processing and transport.</summary>
	public class SimpleQueueScheduler2 : Scheduler
	{
		/// <summary>Constructs a simple queue-based scheduler.</summary>
		/// <param name="topology">Topology of nodes to use.</param>
		public SimpleQueueScheduler2(Topology topology) : base(topology)
		{
			ProcessingQueue = new Queue<ProcessingWorkItem>();
			TransportQueue = new Queue<TransportWorkItem>();
			Status = SchedulerStatus.Offline;
			Verbose = false;
		}

		/// <summary>Ready queue: queue of nodes that are ready to run.</summary>
		private Queue<ProcessingWorkItem> ProcessingQueue { get; set; }

		/// <summary>Ready queue: queue of transport operations.</summary>
		private Queue<TransportWorkItem> TransportQueue { get; set; }

		/// <summary>Indicating whether this <see cref="SimpleQueueScheduler2"/> produces verbose output.</summary>
		public bool Verbose { get; set; }

		/// <summary>Sets up the scheduler, preparing it for the work ahead.</summary>
		public override void Setup()
		{
			base.Setup();

			// let's start the scheduler then
			Status = SchedulerStatus.Starting;

			// register NewFrameEventListener to listen for all newly produced frames
			foreach (var edgeFan in Topology.OutgoingEdges)
			{
				OutputPin outputPin = edgeFan.Key;
				outputPin.NewFrame += new NewFrameEventHandler(NewFrameEventListener);
			}

			// register NodeStatusChangedListener with all nodes, to listen to all node status changes
			foreach (INode node in Topology.Nodes)
				node.StatusChanged += new NodeStatusChangedHandler(NodeStatusChangedListener);

			// start up all nodes by initialising them and updating their status
			foreach (INode node in Topology.Nodes)
			{
				node.Init();
				node.UpdateStatus();
			}

			// let's go over to proper running mode
			Status = SchedulerStatus.Running;
		}

		/// <summary>Main function of a worker thread. This is where the scheduling happens.</summary>
		/// <param name="data">The thread's worker ID, if <see cref="Scheduler.StartWorker"/> is not not overidden.</param>
		protected override void RunWorker(object data)
		{
			try
			{
				// First thread is transport thread
				if (workers[0] == Thread.CurrentThread)
				{
					// Run an infinite loop, looking for work to do
					while (Status == SchedulerStatus.Running)
					{
						TransportWorkItem workItem = null;
						lock (TransportQueue)
						{
							// check if there is work to do (first chance)
							if (TransportQueue.Count == 0)
								Monitor.Wait(TransportQueue, 100);

							// check if there is work to do (second chance)
							if (TransportQueue.Count == 0)
								return;

							workItem = TransportQueue.Dequeue();
						}

						// run the work item
						workItem.Run();
					}
				}
				else // normal worker threads
				{
					// Run an infinite loop, looking for work to do
					while (Status == SchedulerStatus.Running)
					{
						// take first work item from ProcessingQueue queue and run it
						ProcessingWorkItem workItem = null;

						lock (ProcessingQueue)
						{
							// check if there is work to do (first chance)
							if (ProcessingQueue.Count == 0)
							{
								//Console.WriteLine(Thread.CurrentThread.Name + ": waiting for work.");
								Monitor.Wait(ProcessingQueue, 100);
								//Console.WriteLine(Thread.CurrentThread.Name + ": woke up.");
							}

							// check if there is work to do (second chance)
							if (ProcessingQueue.Count == 0)
								return;

							workItem = ProcessingQueue.Dequeue();
						}

						// run the work item
						if (workItem.Node.Status == NodeStatus.Ready)
							workItem.Run();
						else
							workItem.Node.UpdateStatus();
					}
				}

				if(Verbose) Console.WriteLine("{0}: Stopped (Scheduler: {1}).", Thread.CurrentThread.Name, Status);
			}
			catch (Exception e)
			{
				Console.WriteLine("{0}: {1}.", Thread.CurrentThread.Name, e);
			}
		}

		///// <summary>Tries to run the first WorkItem in the given queue.</summary>
		///// <param name="workQueue">Queue of WorkItems.</param>
		//private void TryRunWorkItem(Queue<WorkItem> workQueue)
		//{
		//    // take first work item from ProcessingQueue queue and run it
		//    WorkItem workItem = null;

		//    lock (workQueue)
		//    {
		//        // check if there is work to do (first chance)
		//        if (workQueue.Count == 0)
		//        {
		//            //Console.WriteLine(Thread.CurrentThread.Name + ": waiting for work.");
		//            Monitor.Wait(workQueue, 100);
		//            //Console.WriteLine(Thread.CurrentThread.Name + ": woke up.");
		//        }

		//        // check if there is work to do (second chance)
		//        if (workQueue.Count == 0)
		//            return;

		//        workItem = workQueue.Dequeue();
		//    }

		//    // run the work item
		//    workItem.Run();
		//}

		/// <summary>
		/// Prints the current state.
		/// </summary>
		public void PrintState()
		{
			Console.WriteLine("Scheduler (" + this.Status.ToString() + "):");
			lock (ProcessingQueue)
			{
				Console.WriteLine("  Ready: " + String.Join(", ", ProcessingQueue.Select(workItem => workItem.ToString()).ToArray()));
			}
			Console.WriteLine();
		}

		private void NewFrameEventListener(object sender, NewFrameEventArgs e)
		{
			OutputPin outputPin = (OutputPin)sender;
			//Frame newFrame = e.Frame;

			// schedule transport of frame to all receivers
			foreach (var inputPin in Topology.OutgoingEdges[outputPin])
			{
				// make the input pin listen for new frames from the output pin
				//outputPin.NewFrame += new NewFrameEventHandler(inputPin.NewFrameListener);
				//inputPin.NewFrameListener(this, e);
				lock (TransportQueue)
				{
					TransportQueue.Enqueue(new TransportWorkItem(outputPin, inputPin, e));
					Monitor.Pulse(TransportQueue);
				}
			}
		}

		private void NodeStatusChangedListener(object sender, NodeStatusChangedEventArgs e)
		{
			if (Verbose) Console.WriteLine("{0}: {1} changed: {2} => {3}", Thread.CurrentThread.Name, sender, e.OldStatus, e.NewStatus);
			INode node = (INode)sender;

			// if a node is now ready to run, append it to the ProcessingQueue queue
			if (e.NewStatus == NodeStatus.Ready)
			{
				lock (ProcessingQueue)
				{
					ProcessingQueue.Enqueue(new ProcessingWorkItem(node));
					Monitor.Pulse(ProcessingQueue);
				}
				return;
			}

			// if a node is now waiting for data, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Waiting)
			{
				return;
			}

			// if a node is was ready and is now running, there is nothing we need to do
			if (e.OldStatus == NodeStatus.Ready && e.NewStatus == NodeStatus.Running)
			{
				return;
			}

			// if a node is finished with processing, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Finished)
			{
				return;
			}

			// if a node is now stopped, exit everything
			// TODO: do something more graceful here ...
			if (e.NewStatus == NodeStatus.Stopped)
			{
				Console.Error.WriteLine("Node " + node.Name + " was stopped.");
				Stop();
				return;
			}

			throw new NotImplementedException(String.Format("Cannot handle state transition {0} => {1} of node {2}.", e.OldStatus, e.NewStatus, node));
		}
	}
}
