﻿// $Id: StaticScheduler2.cs 436 2010-10-15 16:26:38Z cr333 $
// Based on a Part II project by Rubin Xu (2010).
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using VideoLib.Frames;
using System.Diagnostics;
using VideoLib.Utils;

namespace VideoLib.Scheduling
{
	/// <summary>Implements a simple GPU-aware static scheduler.</summary>
	/// <remarks>
	/// This scheduler spawns a given number of worker threads, each of which 
	/// deal with a predefined subset (hence static) of all processing nodes.
	/// ALl GPU nodes belong to one worker thread, making the scheduler CUDA-friendly,
	/// Within each worker thread, the next processing node to be served is determined
	/// by FCFS.
	/// 
	/// Worker threads are identified by unique IDs in the range 0..NUM_WORKERS-1, 
	/// Worker 0 is in charge of all GPU nodes. 
	/// 
	/// Nodes' worker affinity can be changed on the fly
	/// </remarks>
	public class StaticScheduler2 : Scheduler
	{
		/// <summary>Constructs a simple static scheduler.</summary>
		/// <param name="topology">Topology of nodes to use.</param>
		public StaticScheduler2(Topology topology) : base(topology)
		{
			Status = SchedulerStatus.Offline;
		}


		private Dictionary<Node, int> NodeAffinity = new Dictionary<Node,int>();
		private HashSet<Node>[] ReadyQueues;
		private Semaphore[] ReadyQueueSemaphores;

		private Thread Daeamon;
		/// <summary>Sets up the scheduler, preparing it for the work ahead.</summary>
		public override void Setup()
		{
			base.Setup();

			try
			{
				// let's start the scheduler then
				Status = SchedulerStatus.Starting;

				// register NewFrameEventListener to listen for all newly produced frames
				foreach (var edgeFan in Topology.OutgoingEdges)
				{
					OutputPin outputPin = edgeFan.Key;
					outputPin.NewFrame += new NewFrameEventHandler(NewFrameEventListener);
				}

				// register NodeStatusChangedListener with all nodes, to listen to all node status changes
				foreach (INode node in Topology.Nodes)
					node.StatusChanged += new NodeStatusChangedHandler(NodeStatusChangedListener);

				// start up all nodes by initialising them (and updating their 
				// status only after the affinity is done, because UpdateStatus()
				// will make some nodes go into Ready, and they need to be put to the 
				// corresponding ready queue.
				foreach (INode node in Topology.Nodes)
				{
					node.Init();
					//node.UpdateStatus();
				}

				// let's go over to proper running mode
				Status = SchedulerStatus.Running;
			}
			catch (Exception)
			{
				throw;
			}
		}

		/// <summary>Dummy implementaion of RunWorker, use RunWorkerThread(Queue) instead.</summary>
		/// <param name="data">The thread's worker ID, if <see cref="Scheduler.StartWorker"/> isnot not overidden.</param>
		protected override void RunWorker(object data)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Create workers threads, also initialise nodes' affinity at this stage        
		/// </summary>
		public override void StartWorkers(int NumWorkers)
		{
			// Assigning nodes' affinity 
			// As an initial guess, distribute the number of nodes evenly 
			// through all worker threads. All GPU nodes go to worker 0.
			int currentAffinity = NumWorkers-1;    
			foreach (Node node in Topology.Nodes)
				if (node is ICudaNode)
				{
					// GPU nodes belong to worker 0
					NodeAffinity[node] = 0;
				}else
				{
					// Otherwise assign affinity in a round-robin manner
					NodeAffinity[node] = currentAffinity;
					currentAffinity = (currentAffinity == 0) ? NumWorkers - 1 : currentAffinity - 1;
				}


			// Initialise Ready queues and AutoResetEvent for each worker
			ReadyQueues = new HashSet<Node>[NumWorkers];
			ReadyQueueSemaphores = new Semaphore[NumWorkers];
			for (int i = 0; i < NumWorkers; i++)
			{
				ReadyQueues[i] = new HashSet<Node>();
				ReadyQueueSemaphores[i] = new Semaphore(0, 100);
			}


			// Run the initial UpdateStatus() after the affinity is set up
			foreach (INode node in Topology.Nodes)
				node.UpdateStatus();

			// Spawn worker threads
			for (int i = 1; i <= NumWorkers; i++)
			{
				Thread worker = new Thread(RunWorkerThread) { Name = String.Format("Worker{0}", i) };
				worker.Start(i-1);
				workers.Add(worker);
			}

			//Spawn a manager thread which will adjust processor
			//allocation once enough profiling data is collected.
			Daeamon = new Thread(AffinitySchedulingDaemon);
			Daeamon.Start();
		}

		/// <summary>
		/// Adjust a node's affinity on the fly.
		/// </summary>
		/// <param name="node">target node </param>
		/// <param name="Affinity">new affinity for that node</param>
		public void ChangeAffinity(Node node, int Affinity)
		{
			if (Affinity >= workers.Count)
				throw new InvalidOperationException("Invalid affinity for node " + node.ToString());

			lock(NodeAffinity)
			{
				NodeAffinity[node] = Affinity;
			}
		}

		/// <summary>Main function of a worker thread</summary>
		protected void RunWorkerThread(Object data)
		{
			//Try to lock managed thread to one native thread
			Thread.BeginThreadAffinity();

			// Run an infinite loop, looking for work to do
			int workerID = (Int32)data;

			HashSet<Node> ready = ReadyQueues[workerID];
			Node node;

			Dictionary<Node, long> LastServedTime = new Dictionary<Node, long>();
			while (Status == SchedulerStatus.Running)
			{
				try
				{
					ReadyQueueSemaphores[workerID].WaitOne(1000);
				}catch
				{
					return;
				}


				lock (ready)
				{
					if ((ready.Count != 0) && (Status == SchedulerStatus.Running))
					{
						//Find the node which hasn't been served for the longest time
						long T = Int64.MaxValue;
						node = null;
						foreach(Node n in ready)
							if (node == null)
							{
								node = n;
							}
							else 
							if (LastServedTime.ContainsKey(n) && LastServedTime[n] < T)
							{
								T = LastServedTime[n];
								node = n;
							}
						ready.Remove(node);
					}else
					{
						node = null;
					}
				}

				if (node != null)
				{
					LastServedTime[node] = Stopwatch.GetTimestamp();

					Debug.WriteLine(String.Format("{0}: Node<{1}> processing", Thread.CurrentThread.Name, node.ToString()));
					Profiler.MarkStarting(node);
					node.PreProcess();
					node.Process();
					node.PostProcess();
					Profiler.MarkEnding(node);
				}
			}

			Console.WriteLine(Thread.CurrentThread.Name + ": Stopped.");

			Thread.EndThreadAffinity();
		}

		// Propagate newFrame event to all inputPins
		private void NewFrameEventListener(object sender, NewFrameEventArgs e)
		{
			OutputPin outputPin = (OutputPin)sender;
			foreach (var inputPin in Topology.OutgoingEdges[outputPin])
			{
				inputPin.NewFrameListener(outputPin, e);
				Debug.WriteLine(String.Format("Transport: {0} => {1}", e.Frame.GetType().Name, outputPin));
			}
		}

		private void NodeStatusChangedListener(object sender, NodeStatusChangedEventArgs e)
		{
			Debug.WriteLine(String.Format("{0}: {1} changed: {2} => {3}", Thread.CurrentThread.Name, sender, e.OldStatus, e.NewStatus));
			Node node = (Node)sender;

			// if a node is now ready to run, signal it.
			if (e.NewStatus == NodeStatus.Ready)
			{
				int affinity;
				lock(NodeAffinity)
				{
					affinity = NodeAffinity[node];
				}

				HashSet<Node>  ReadyQ = ReadyQueues[affinity];
				lock (ReadyQ)
				{
					ReadyQ.Add(node);
				}
				ReadyQueueSemaphores[affinity].Release();
				return;
			}

			// if a node is now waiting for data, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Waiting)
			{
				return;
			}

			// if a node is was ready and is now running, there is nothing we need to do
			if (e.OldStatus == NodeStatus.Ready && e.NewStatus == NodeStatus.Running)
			{
				return;
			}

			// if a node is finished with processing, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Finished)
			{
				return;
			}

			// if a node is now stopped, exit everything
			// TODO: do something more graceful here ...
			if (e.NewStatus == NodeStatus.Stopped)
			{
				foreach (Thread worker in workers)
					worker.Interrupt();
				Console.Error.WriteLine("Node " + node.Name + " was stopped.");
				Stop();
				return;
			}

			//throw new NotImplementedException(String.Format("Cannot handle state transition {0} => {1} of node {2}.", e.OldStatus, e.NewStatus, node));
		}

		/// <summary>
		/// Stops the scheduler.
		/// </summary>
		public override void Stop()
		{
			if (Daeamon != null)
				Daeamon.Interrupt();
			base.Stop();
		}

		/// <summary>
		/// Daemon thread that does the clever scheduling (adjust node affinity)
		/// once the scheduler has collected enough profiling data.
		/// </summary>
		private void AffinitySchedulingDaemon()
		{
			try
			{
				//Wait for 7 sec to collect profiling data
				Thread.Sleep(7 * 1000);
			}
			catch (System.Exception)
			{
				//In case the pipeline is stopped before 7 sec.
				return;
			}
			Profiler.PrepareResults();
			
			//Get profiling results
			//Only contain all CPU nodes. Do not adjust GPU affinity 
			Dictionary<Object, double> Costs = new Dictionary<Object, double>();
			foreach(Object obj in Profiler.GetProfilingTargets())
				if (!(obj is ICudaNode))
					Costs[obj] = Profiler.GetResult(obj).RunTimeMean;

			// Get the number of worker threads from ReadyQueues (a little strange..)
			// As worker 0 is for GPU only, the actual available threads is one less.
			int NumCPUThreads = ReadyQueues.Length - 1;
			AbstractPartitionSolver ps = new CompleteGreedyAlgorithmPartitionSolver();
			Dictionary<Object, int> optimalAffinity = ps.FindOptimalAffinity(NumCPUThreads, Costs, new HashSet<Object>());

			//Debug printouts.
			Console.WriteLine("Static Scheduler - Initial Affinity:");
			for (int i = 0; i < NumCPUThreads + 1; i++)
			{
				double total = 0.0;
				Console.Write("Worker " + i.ToString() + ": ");
				foreach (KeyValuePair<Node, int> pr in NodeAffinity)
					if (pr.Value == i)
					{
						if (Costs.ContainsKey(pr.Key))
							total += Costs[pr.Key];
						Console.Write(String.Format("{0}({1,5:F2}ms) ", pr.Key.ToString(), (Costs.ContainsKey(pr.Key) ? Costs[pr.Key] * 1000.0 : 0.0)));
					}
				Console.WriteLine(" Total: {0,5:F2}ms", total * 1000.0);
			}

			//Update affinity
			lock(NodeAffinity)
			{
				// Resultant affinity range is 0..NumThreads-1, but we want to
				// allocate them to worker_1..worker_NumThreads
				foreach (KeyValuePair<Object, int> pr in optimalAffinity)
					NodeAffinity[(Node)pr.Key] = pr.Value + 1;
			}

			//Debug printouts.
			Console.WriteLine("Static Scheduler - Modified Affinity:");
			for (int i = 0; i < NumCPUThreads + 1; i++)
			{
				double total = 0.0;
				Console.Write("Worker " + i.ToString()+ ": ");
				foreach (KeyValuePair<Node, int> pr in NodeAffinity)
					if(pr.Value == i)
					{
						if (Costs.ContainsKey(pr.Key))
							total += Costs[pr.Key];
						Console.Write(String.Format("{0}({1,5:F2}ms) ", pr.Key.ToString(), (Costs.ContainsKey(pr.Key) ? Costs[pr.Key]*1000.0 : 0.0)));
					}
				Console.WriteLine(" Total: {0,5:F2}ms", total*1000.0);
			}
		}

	}
}
