﻿// $Id: SingleImageSource.cs 337 2010-08-17 18:08:27Z cr333 $
// Based on a Part II project by Rubin Xu (2010).
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.Frames;
using System.IO;
using System.Drawing;
using System.Threading;

namespace VideoLib.Sources
{
	/// <summary>
	/// Loads a single images from disk and push it through the pipeline continuously.
	/// Slightly more efficient than ImageSequenceSource with loop, as the image is loaded from disk once only.
	/// </summary>
	public class SingleImageSource : Node
	{
		/// <summary>Filename of the image.</summary>
		public string Filename { get; set; }

		/// <summary> The BitmapFrame object corresponding to current image </summary>
		private BitmapFrame ImageFrame = null;

		private int CurrentSequenceNumber;
		/// <summary>The frame rate with which the video frames will be output.</summary>
		public float FPS { get; set; }

		/// <summary>Human-friendly node name.</summary>
		public override string Name
		{
			get { return String.Format("{0}({1})", base.Name, Path.GetFileName(Filename)); }
		}

		/// <summary>Time of the last frame, to control FPS.</summary>
		private DateTime lastFrameTime = DateTime.Now;
		/// <summary>Time of the first frame, to adjust long term FPS.</summary>
		private DateTime FirstFrameTime;

		/// <summary>Constructs a new image sequence source.</summary>
		/// <param name="filename">Filename for the image</param>
		public SingleImageSource(string filename)
			: this()
		{
			Filename = filename;
			LoadImageFromDisk(Filename);
		}

		/// <summary>Creates a new source based on a single cached input image.</summary>
		public SingleImageSource() 
			: base()
		{
			// Default settings
			FPS = 25;
			CurrentSequenceNumber = 0;
			AddOutputPin("out");
		}

		/// <summary>Prepares the image sequence source.</summary>
		public override void Init()
		{
			CurrentSequenceNumber = 0;
		}

		/// <summary>Create a Bitmap object given the image's filename </summary>
		/// <param name="filename"></param>
		/// <returns></returns>
		public void LoadImageFromDisk(String filename)
		{
			Filename = filename;
			Bitmap CurrentBitmap = new Bitmap(filename);
			ImageFrame = new BitmapFrame(CurrentBitmap);
		}

		/// <summary>Processes a frame.</summary>
		public override void Process()
		{
			Frame frame = null;
			if (CurrentSequenceNumber == 0)
				FirstFrameTime = DateTime.Now;

			// Create new BitmapFrame 
			ImageFrame.copyTo(ref frame);
			frame.SequenceNumber = CurrentSequenceNumber;

			// push new frame, updating internal state
			PushFrame("out", frame);

			// sleep to get frame rate right
			Thread.Sleep((int)Math.Max(0, 1000 / FPS - (DateTime.Now.Ticks - lastFrameTime.Ticks) / 1e4));
			lastFrameTime = DateTime.Now;

			int ExpectedSequenceNumber = (int)Math.Round((lastFrameTime.Ticks - FirstFrameTime.Ticks) * FPS / 1e7);
			// Check if we are behind the expected frame rate, originally will emit 
			// as many frames as required to compensate for the correct frame rate,
			// but this will eat up all memory very quickly if the frame rate is set too high,
			// So as a workaround only emit one extra frame if we are behind
			if (CurrentSequenceNumber < ExpectedSequenceNumber)
			{
				CurrentSequenceNumber++;
				frame = null;
				ImageFrame.copyTo(ref frame);
				frame.SequenceNumber = CurrentSequenceNumber;
				PushFrame("out", frame);
				lastFrameTime = DateTime.Now;
			}
			CurrentSequenceNumber++;
			/*
			for (int i = CurrentSequenceNumber + 1; i <= ExpectedSequenceNumber; i++)
			{
				//Push additional frames if we are behind 
				frame = null;
				ImageFrame.copyTo(ref frame);
				frame.SequenceNumber = i;
				PushFrame("out", frame);
			lastFrameTime = DateTime.Now;
			}
			CurrentSequenceNumber = ExpectedSequenceNumber + 1;
			 */
		}
	}
}
