﻿// $Id: Topology.cs 550 2012-02-15 18:07:05Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace VideoLib
{
	/// <summary>A class that encapsulates a topology of <see cref="Node"/>s.</summary>
	public class Topology
	{
		/// <summary>Initialises a new instance of the <see cref="Topology"/> class.</summary>
		public Topology()
		{
			Clear();
		}

		// implement singleton stuff for easy-to-use default topology
		/// <summary>Gets the main topology instance (singleton).</summary>
		public static Topology MainInstance { get { return Nested.instance; } }

		class Nested
		{
			// Explicit static constructor to tell C# compiler
			// not to mark type as beforefieldinit
			static Nested() { }
			internal static readonly Topology instance = new Topology();
		}

		/// <summary>Gets the list of all <see cref="INode"/>s.</summary>
		public List<INode> Nodes { get; protected set; }

		/// <summary>Gets all edges, organised by destination <see cref="InputPin"/>.</summary>
		/// <value>The incoming edges.</value>
		public Dictionary<InputPin, List<OutputPin>> IncomingEdges { get; protected set; }

		/// <summary>Gets all edges, organised by source <see cref="OutputPin"/>.</summary>
		/// <value>The outgoing edges.</value>
		public Dictionary<OutputPin, List<InputPin>> OutgoingEdges { get; protected set; }

		/// <summary>Signals that the topology has finished.</summary>
		public AutoResetEvent Finished { get { return finished; } }
		private AutoResetEvent finished = new AutoResetEvent(false);

		/// <summary>Adds a node to the topology without connecting it.</summary>
		/// <param name="node">The node to be added.</param>
		public void AddNode(INode node)
		{
			Nodes.Add(node);
		}

		/// <summary>Removes a node from the topology, unlinking all edges.</summary>
		/// <param name="node">The node to be removed.</param>
		public void RemoveNode(INode node)
		{
			// remove all edges to and from the node
			foreach (OutputPin outputPin in node.OutputPins.Values)
				OutgoingEdges.Remove(outputPin);
			foreach (InputPin inputPin in node.InputPins.Values)
				IncomingEdges.Remove(inputPin);

			// finally remove the node
			Nodes.Remove(node);
		}

		/// <summary>Resets the current topology.</summary>
		public void Clear()
		{
			Nodes = new List<INode>();
			IncomingEdges = new Dictionary<InputPin, List<OutputPin>>();
			OutgoingEdges = new Dictionary<OutputPin, List<InputPin>>();
		}

		/// <summary>Connects an output pin of one node with an input pin on another node.</summary>
		/// <param name="outputPin">Output pin of producer node.</param>
		/// <param name="inputPin">Input pin of consumer node.</param>
		public void Connect(OutputPin outputPin, InputPin inputPin)
		{
			// ensure that lists exist
			if (!OutgoingEdges.ContainsKey(outputPin)) OutgoingEdges[outputPin] = new List<InputPin>();
			if (!IncomingEdges.ContainsKey(inputPin))  IncomingEdges[inputPin]  = new List<OutputPin>();

			// add outgoing edge
			OutgoingEdges[outputPin].Add(inputPin);

			// add incoming edge
			IncomingEdges[inputPin].Add(outputPin);
		}

		/// <summary>Connects default outputs to default inputs in a chain of nodes.</summary>
		/// <param name="nodeChain">List of nodes or IO pins to connect in a chain.</param>
		/// <returns>The last node, if it is a node, <c>null</c> otherwise.</returns>
		/// <remarks>The first item can be an output pin, and the last one an input pin.</remarks>
		public INode Connect(params Object[] nodeChain)
		{
			// skip if not sufficiently many nodes
			if (nodeChain.Length < 1) return null;
			if (nodeChain.Length < 2) return nodeChain[0] as INode;

			for (int sourceIndex = 0; sourceIndex + 1 < nodeChain.Length; sourceIndex++)
			{
				// convert current item to an output pin
				OutputPin source = GetAsOutputPin(nodeChain[sourceIndex]);

				// convert next item to an input pin
				InputPin sink = GetAsInputPin(nodeChain[sourceIndex + 1]);

				// check if the pins are valid
				if (source == null) throw new ArgumentException(String.Format("Could not find output pin of {0} (parameter {1}).", nodeChain[sourceIndex], sourceIndex + 1));
				if (sink == null)   throw new ArgumentException(String.Format("Could not find input pin of {0} (parameter {1}).", nodeChain[sourceIndex + 1], sourceIndex + 2));

				// connect together
				Connect(source, sink);
			}

			return nodeChain[nodeChain.Length - 1] as INode;
		}

		/// <summary>Connects the first item to the following items.</summary>
		/// <remarks>Items can be IO pins, if appropriate.</remarks>
		/// <param name="nodeChain">List of nodes or IO pins to connect in a chain.</param>
		public void ConnectFan(params Object[] nodeChain)
		{
			// skip if not sufficiently many nodes
			if (nodeChain.Length < 2) return;

			// convert first item in list to an output pin
			OutputPin source = GetAsOutputPin(nodeChain[0]);

			// check if valid
			if (source == null)
				throw new ArgumentException(String.Format("Could not find output pin of {0} (first parameter).", nodeChain[0]));

			for (int sinkIndex = 1; sinkIndex < nodeChain.Length; sinkIndex++)
			{
				// convert next item to an input pin
				InputPin sink = GetAsInputPin(nodeChain[sinkIndex]);

				// check if valid
				if (sink == null)
					throw new ArgumentException(String.Format("Could not find input pin of {0} (parameter {1}).", nodeChain[sinkIndex], sinkIndex + 1));

				// connect together
				Connect(source, sink);
			}
		}

		/// <summary>Converts an object to an <see cref="InputPin"/> if possible.</summary>
		/// <param name="input">An input pin or node with default input pin.</param>
		/// <returns>An instance of <see cref="InputPin"/>, or null if the conversion failed.</returns>
		private static InputPin GetAsInputPin(Object input)
		{
			// first check if it's an input pin
			var inputPin = input as InputPin;
			if (inputPin != null)
				return inputPin;

			// then check for default input pin on node
			var node = input as INode;
			if (node != null && node.InputPins.ContainsKey("in"))
				return node.InputPins["in"];

			// failed to find input pin
			return null;
		}

		/// <summary>Converts an object to an <see cref="OutputPin"/> if possible.</summary>
		/// <param name="output">An output pin or node with default output pin.</param>
		/// <returns>An instance of <see cref="OutputPin"/>, or null if the conversion failed.</returns>
		private static OutputPin GetAsOutputPin(Object output)
		{
			// first check if it's an output pin
			var outputPin = output as OutputPin;
			if (outputPin != null)
				return outputPin;

			// then check for default output pin on node
			var node = output as INode;
			if (node != null && node.OutputPins.ContainsKey("out"))
				return node.OutputPins["out"];

			// failed to find output pin
			return null;
		}


		private int runningNodes;

		/// <summary>
		/// Waits for all nodes in the topology to have finished.
		/// </summary>
		public void Join()
		{
			// Number of nodes in this topology.
			runningNodes = Nodes.Count;

			// Wait until they're all finished.
			Finished.WaitOne();
		}

		/// <summary>
		/// Call this function to signal that a node has finished processing.
		/// </summary>
		/// <remarks>Be careful to call this function exactly once for each node.</remarks>
		public void NodeFinished()
		{
			// Raise the 'Finished' event when the last node has finished running.
			if (Interlocked.Decrement(ref runningNodes) == 0)
				Finished.Set();
		}

		/// <summary>
		/// Prints the current state.
		/// </summary>
		public void PrintState()
		{
			Console.WriteLine("Topology:");
			Console.WriteLine("  Nodes:");
			Nodes.ForEach(delegate(INode node) { Console.WriteLine("    - " + node.ToString()); });
			Console.WriteLine();

			// print info about edges in topology
			Console.WriteLine("  Edges:");
			foreach (OutputPin outputPin in OutgoingEdges.Keys)
				OutgoingEdges[outputPin].ForEach(inputPin => Console.WriteLine("    - " + outputPin + " => " + inputPin));
			Console.WriteLine();
		}

		internal PinStatus UpdateStatus(OutputPin outputPin)
		{
			// no edge leaving this pin? => always ready
			if (!OutgoingEdges.Keys.Contains(outputPin))
				return PinStatus.Ready;

			// block this output pin if any downstream node's input pin is blocked
			foreach (InputPin inputPin in OutgoingEdges[outputPin])
				if (inputPin.Status == PinStatus.Blocked)
					return PinStatus.Blocked;

			// nothing blocked? ready to run!
			return PinStatus.Ready;
		}

		/// <summary>Checks the connections between nodes and warns if there are any issues.</summary>
		public void CheckConnections()
		{
			foreach (var node in Nodes)
			{
				foreach (var inputPin in node.InputPins)
				{
					if (IncomingEdges.ContainsKey(inputPin.Value) == false)
					{
						Console.WriteLine(String.Format("Warning: Input pin {0}.{1} is not connected.", node.Name, inputPin.Value.Name));
					}
					else if(IncomingEdges[inputPin.Value].Count > 1)
					{
						Console.WriteLine(String.Format("Warning: Input pin {0}.{1} has more than one input ({2} inputs).", node.Name, inputPin.Value.Name, IncomingEdges[inputPin.Value].Count));
					}
				}
			}
		}

		#region Export to DOT graph

		private static string FormatNodeAsDotRecord(INode node)
		{
			// normal nodes are white, CUDA is green, OpenGL blue
			var colour = "white";
			if (node.GetType().GetInterfaces().Select(x => x.Name).Contains("ICudaNode"))   colour = "palegreen";
			if (node.GetType().GetInterfaces().Select(x => x.Name).Contains("IOpenGLNode")) colour = "lightskyblue";

			var inputs  = String.Join(" | ", node.InputPins.Values.OrderBy(x => x.Name).Select(x => string.Format("<{0}> {0}", x.Name)).ToArray());
			var outputs = String.Join(" | ", node.OutputPins.Values.OrderBy(x => x.Name).Select(x => string.Format("<{0}> {0}", x.Name)).ToArray());

			return string.Format("{0} [shape = record, style = filled, fillcolor = {1}, label = \"<node> {2} |{{ {{ {3} }} | {{ {4} }} }}\"];",
				node.GetHashCode(), colour, node.ToString(), inputs, outputs);
		}

		/// <summary>Exports the topology as a dot graph for visualisation.</summary>
		/// <returns>DOT source file</returns>
		public string ExportToDotGraph()
		{
			var sb = new StringBuilder();

			// write header
			sb.AppendLine("digraph g {");
			sb.AppendLine("graph [rankdir = LR]");
			sb.AppendLine("node [fontsize = 10, fontname = Verdana]");
			sb.AppendLine("edge []");
			sb.AppendLine();

			// output sources first
			sb.AppendLine("// sources");
			sb.AppendLine("{ rank = same");
			foreach (var node in Nodes.Where(x => x.InputPins.Count == 0))
				sb.AppendLine(FormatNodeAsDotRecord(node));
			sb.AppendLine("}");
			sb.AppendLine();

			// then output sinks
			sb.AppendLine("// sinks");
			sb.AppendLine("{ rank = same");
			foreach (var node in Nodes.Where(x => x.InputPins.Count != 0 && x.OutputPins.Count == 0))
				sb.AppendLine(FormatNodeAsDotRecord(node));
			sb.AppendLine("}");
			sb.AppendLine();

			// finally output filters
			sb.AppendLine("// filters");
			foreach (var node in Nodes.Where(x => x.InputPins.Count != 0 && x.OutputPins.Count != 0))
				sb.AppendLine(FormatNodeAsDotRecord(node));
			sb.AppendLine();

			// output edges
			sb.AppendLine("// edges");
			foreach (var edgeFan in OutgoingEdges)
				foreach (var edge in edgeFan.Value)
					sb.AppendLine(string.Format("{0}:{1} -> {2}:{3};",
						edgeFan.Key.Node.GetHashCode(), edgeFan.Key.Name,
						edge.Node.GetHashCode(), edge.Name));

			sb.AppendLine("}");
			return sb.ToString();
		}

		#endregion Export to DOT graph
	}
}
