﻿// $Id: MemoryAllocator.cs 436 2010-10-15 16:26:38Z cr333 $
// Based on a Part II project by Rubin Xu (2010).
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace VideoLib.Utils
{
	/// <summary>
	/// Base class for memory allocators that provide byte arrays.
	/// </summary>
	public interface MemoryAllocator
	{
		/// <summary>
		/// Allocates an array of bytes of the requested size.
		/// </summary>
		/// <param name="size">Size of the requested byte array</param>
		/// <returns>An array of bytes of the requested size.</returns>
		byte[] AllocArray(int size);

		/// <summary>
		/// Frees the given array of bytes.
		/// </summary>
		/// <param name="array">The array of bytes to be freed.</param>
		void FreeArray(byte[] array);

		/// <summary>
		/// Frees all allocated memory.
		/// </summary>
		void FreeAll();
	}


	/// <summary>
	/// Allocates memory using 'new' and leaves deallocation to the grabage collector.
	/// </summary>
	public class BuiltinAllocator : MemoryAllocator
	{
		/// <summary>
		/// Allocates an array of bytes of the requested size, using 'new'.
		/// </summary>
		/// <param name="size">Size of the requested byte array</param>
		/// <returns>An array of bytes of the requested size.</returns>
		public byte[] AllocArray(int size)
		{
			return new byte[size];
		}

		/// <summary>
		/// Frees the given array of bytes.
		/// </summary>
		/// <param name="array">The array of bytes to be freed.</param>
		public void FreeArray(byte[] array)
		{
			// nothing to do here, as the garbage collector takes care of it
		}

		/// <summary>
		/// Frees all allocated memory.
		/// </summary>
		public void FreeAll()
		{
			// nothing to do here, as the garbage collector takes care of it
		}
	}


	/// <summary>
	/// Rubin's memory pool allocator (without everything being static).
	/// </summary>
	public class MemoryPoolAllocator : MemoryAllocator
	{
		private Dictionary<int, LinkedList<WeakReference>> pool = new Dictionary<int, LinkedList<WeakReference>>();
		private Object lockobj = new Object();

		/// <summary>
		/// Returns an array of bytes of the requested size from the memory pool,
		/// or create a new one.
		/// </summary>
		/// <param name="size">Size of the requested byte array</param>
		/// <returns>An array of bytes of the requested size.</returns>
		public byte[] AllocArray(int size)
		{
			byte[] newarray = null;
			GC.WaitForPendingFinalizers();

			lock (lockobj)
			{
				// make sure a pool of the given size exists
				if (!pool.ContainsKey(size))
					pool[size] = new LinkedList<WeakReference>();

				// try to find a byte array in the pool
				LinkedList<WeakReference> subpool = pool[size];
				while ((subpool.Count > 0) && (newarray == null))
				{
					newarray = (byte[])subpool.First().Target;
					subpool.RemoveFirst();
				}

				// allocate memory directly if the pool is empty
				if (newarray == null)
					newarray = new byte[size];
			}

			return newarray;
		}

		/// <summary>
		/// Frees the given array of bytes by putting it back in the pool.
		/// </summary>
		/// <param name="array">The array of bytes to be freed.</param>
		public void FreeArray(byte[] array)
		{
			LinkedList<WeakReference> subpool = pool[array.Length];
			lock (lockobj)
			{
				subpool.AddLast(new WeakReference(array));
			}
		}

		/// <summary>
		/// Frees all allocated memory.
		/// </summary>
		public void FreeAll()
		{
			foreach (var pr in pool)
				pr.Value.Clear();
			pool.Clear();
		}

		/// <summary>
		/// Clean up the memory pools. This will not delete any data,
		/// just the pools with weak references.
		/// </summary>
		~MemoryPoolAllocator()
		{
			FreeAll();
		}
	}


	/// <summary>
	/// Christian's reimplemented memory pool allocator.
	/// </summary>
	/// <remarks>
	/// Similiar to Rubin's memory pool allocator, but with a per-pool lock instead of a global lock.
	/// </remarks>
	public class MultiMemoryPoolAllocator : MemoryAllocator
	{
		private class SingleSizePool
		{
			private LinkedList<WeakReference> pool = new LinkedList<WeakReference>();
			private Object mutex = new Object();
			public int Size { get; private set; }

			public SingleSizePool(int size)
			{
				Size = size;
			}

			/// <summary>
			/// Returns an array of the size configured for this pool.
			/// </summary>
			public byte[] AllocArray()
			{
				byte[] newarray = null;
				GC.WaitForPendingFinalizers();

				lock (mutex)
				{
					// try to find a byte array in the pool
					while ((pool.Count > 0) && (newarray == null))
					{
						newarray = (byte[])pool.First().Target;
						pool.RemoveFirst();
					}

					// allocate memory directly if the pool is empty
					if (newarray == null)
						newarray = new byte[Size];
				}

				return newarray;
			}

			/// <summary>
			/// Frees the given array of bytes by putting it back in the pool.
			/// </summary>
			/// <param name="array">The array of bytes to be freed.</param>
			public void FreeArray(byte[] array)
			{
				lock (mutex)
				{
					pool.AddLast(new WeakReference(array));
				}
			}

			/// <summary>
			/// Frees all allocated memory.
			/// </summary>
			public void FreeAll()
			{
				pool.Clear();
			}

			/// <summary>
			/// Cleans up the memory pool.
			/// </summary>
			/// <remarks>
			/// This will not delete any data, just the weak references to it.
			/// </remarks>
			~SingleSizePool()
			{
				FreeAll();
			}
		}

		private Dictionary<int, SingleSizePool> pools = new Dictionary<int, SingleSizePool>();
		private Object mutex = new Object();

		/// <summary>
		/// Returns an array of bytes of the requested size from the memory pool,
		/// or create a new one.
		/// </summary>
		/// <param name="size">Size of the requested byte array</param>
		/// <returns>An array of bytes of the requested size.</returns>
		public byte[] AllocArray(int size)
		{
			//GC.WaitForPendingFinalizers();

			// make sure a pool of the given size exists
			if (!pools.ContainsKey(size))
			{
				lock (mutex)
				{
					pools[size] = new SingleSizePool(size);
				}
			}

			return pools[size].AllocArray();
		}

		/// <summary>
		/// Frees the given array of bytes by putting it back in the pool.
		/// </summary>
		/// <param name="array">The array of bytes to be freed.</param>
		public void FreeArray(byte[] array)
		{
			if (!pools.ContainsKey(array.Length))
			{
				lock (mutex)
				{
					pools[array.Length] = new SingleSizePool(array.Length);
				}
			}
			pools[array.Length].FreeArray(array);
		}

		/// <summary>
		/// Frees all allocated memory.
		/// </summary>
		public void FreeAll()
		{
			foreach (var pool in pools)
				pool.Value.FreeAll();
			pools.Clear();
		}

		/// <summary>
		/// Cleans up all the memory pools.
		/// </summary>
		/// <remarks>
		/// This will not delete any data, just the pools with weak references.
		/// </remarks>
		~MultiMemoryPoolAllocator()
		{
			FreeAll();
		}
	}
}
